from tradingenv.transmitter import (
    AsynchronousTransmitter,
    AbstractTransmitter,
    Transmitter,
    Folds,
)
from tradingenv.events import IEvent
from datetime import datetime, timezone
import time as time_
import numpy as np
import pandas as pd
import pytest
from collections import Counter


class MarketEvent(IEvent):
    def __init__(self, time):
        self.time = time


class TestEmptyTransmitter:
    def test_parent_class_is_Transmitter(self):
        assert issubclass(AsynchronousTransmitter, AbstractTransmitter)

    def test_now(self):
        transmitter = AsynchronousTransmitter()
        time_.sleep(0.002)
        now = datetime.utcnow()#.replace(tzinfo=timezone.utc)
        assert (transmitter._now() - now).total_seconds() < 0.001

    def test_create_partitions(self):
        transmitter = AsynchronousTransmitter()
        transmitter._create_partitions()  # does not raise error


class TestTransmitter:
    def test_parent_class_is_Transmitter(self):
        assert issubclass(Transmitter, AbstractTransmitter)

    def test_reset_create_partitions_calling_reset_for_the_first_time(self):
        transmitter = Transmitter(timesteps=[datetime.now()])
        assert transmitter._partition_nonlatent is None
        transmitter._reset()
        assert transmitter._partition_nonlatent is not None

    def test_sampling_span(self):
        # There are 10 dates
        # ##########
        # We cannot sample the last 4 as episode_length is 5
        # ######
        # Gamma is 1 / (1 - sampling_span) = 0.5
        # Therefore the probabilities of the first six dates should be:
        #   p = 0.5 ** np.arange(6)
        #   p /= p.sum()
        # That is [0.015625, 0.03125, 0.0625, 0.125, 0.25, 0.50]
        start_dates = list(pd.date_range('2023', periods=10, freq='D'))
        events = [MarketEvent(t) for t in start_dates]
        transmitter = Transmitter(timesteps=start_dates)
        transmitter.add_events(events)

        start_dates = list()
        np.random.seed(0)
        for i in range(10000):
            transmitter._reset(episode_length=5, sampling_span=2)
            start_dates.append(transmitter._start_date)
        c = Counter(start_dates)
        c = {k: c[k] / sum(c.values()) for k in sorted(c)}

        actual_dates = list(c.keys())
        expected_dates = list(pd.date_range('2023', periods=6, freq='D'))
        assert actual_dates == expected_dates

        actual_freq = list(c.values())
        expected_freq = [0.015625, 0.03125, 0.0625, 0.125, 0.25, 0.50]
        np.testing.assert_allclose(actual_freq, expected_freq, rtol=0.2)

    def test_now_origin(self):
        # Origin is first date less avg diff
        transmitter = Transmitter(timesteps=[datetime.now()])
        transmitter._create_partitions()
        transmitter._reset()
        assert transmitter._now() is np.nan

    def test_split_raises_if_periods_have_more_then_two_timestamps(self):
        with pytest.raises(ValueError):
            Transmitter(
                timesteps=[datetime.now()],
                folds={"training-set": [datetime(2019, 1, 5), datetime(2019, 1, 1)]},
            )

    def test_split_raises_if_periods_have_less_then_two_timestamps(self):
        with pytest.raises(ValueError):
            Transmitter(
                timesteps=[datetime.now()],
                folds={"training-set": [datetime(2019, 1, 1)]},
            )

    def test_split_raises_if_start_is_older_than_end(self):
        with pytest.raises(ValueError):
            Transmitter(
                timesteps=[datetime.now()],
                folds={"training-set": [datetime(2019, 1, 5), datetime(2019, 1, 1)]},
            )

    @pytest.mark.skip
    def test_split_raises_if_overlapping_periods_edge(self):
        with pytest.raises(ValueError):
            Transmitter(
                timesteps=[datetime.now()],
                folds={
                    "training-set": [datetime(2019, 1, 1), datetime(2019, 1, 5)],
                    "test-set": [datetime(2019, 1, 5), datetime(2019, 1, 10)],
                },
            )

    @pytest.mark.skip
    def test_split_raises_if_overlapping_periods(self):
        with pytest.raises(ValueError):
            Transmitter(
                timesteps=[datetime.now()],
                folds={
                    "training-set": [datetime(2019, 1, 1), datetime(2019, 1, 5)],
                    "test-set": [datetime(2019, 1, 4), datetime(2019, 1, 10)],
                },
            )

    @pytest.mark.skip
    def test_split_does_not_raises_if_overlapping_datetime_min(self):
        with pytest.raises(ValueError):
            Transmitter(
                timesteps=[datetime.now()],
                folds={
                    "training-set": [datetime.min, datetime(2019, 1, 5)],
                    "test-set": [datetime(2019, 1, 4), datetime(2019, 1, 10)],
                },
            )

    def test_walk_forward_idx_have_same_length(self):
        for train_size in range(1, 5):
            for test_size in range(1, 5):
                transmitter = Transmitter(
                    timesteps=pd.date_range("2019-01-01", periods=20, freq="B")
                )
                folds = transmitter.walk_forward(
                    train_size=train_size, test_size=test_size
                )
                assert len(folds.train_start) == len(folds.train_end)
                assert len(folds.train_start) == len(folds.test_start)
                assert len(folds.train_start) == len(folds.test_end)

    @pytest.mark.parametrize(
        argnames=[
            "nr_obs",
            "expected_train_start",
            "expected_train_end",
            "expected_test_start",
            "expected_test_end",
        ],
        ids=[str(i) for i in range(1, 20)],  # nr_obs
        argvalues=[
            (1, [], [], [], []),
            (2, [], [], [], []),
            (3, [], [], [], []),
            (4, [], [], [], []),
            (5, [], [], [], []),
            (6, [], [], [], []),
            (7, [], [], [], []),
            (8, [0], [4], [5], [7]),
            (9, [0], [4], [5], [7]),
            (10, [0], [4], [5], [7]),
            (11, [0, 3], [4, 7], [5, 8], [7, 10]),
            (12, [0, 3], [4, 7], [5, 8], [7, 10]),
            (13, [0, 3], [4, 7], [5, 8], [7, 10]),
            (14, [0, 3, 6], [4, 7, 10], [5, 8, 11], [7, 10, 13]),
            (15, [0, 3, 6], [4, 7, 10], [5, 8, 11], [7, 10, 13]),
            (16, [0, 3, 6], [4, 7, 10], [5, 8, 11], [7, 10, 13]),
            (17, [0, 3, 6, 9], [4, 7, 10, 13], [5, 8, 11, 14], [7, 10, 13, 16]),
            (18, [0, 3, 6, 9], [4, 7, 10, 13], [5, 8, 11, 14], [7, 10, 13, 16]),
            (19, [0, 3, 6, 9], [4, 7, 10, 13], [5, 8, 11, 14], [7, 10, 13, 16]),
        ],
    )
    def test_split(
        self,
        nr_obs,
        expected_train_start,
        expected_train_end,
        expected_test_start,
        expected_test_end,
    ):
        transmitter = Transmitter(
            timesteps=pd.date_range("2019-01-01", periods=nr_obs, freq="B")
        )
        folds = transmitter.walk_forward(train_size=5, test_size=3)
        np.testing.assert_equal(folds.train_start, expected_train_start)
        np.testing.assert_equal(folds.train_end, expected_train_end)
        np.testing.assert_equal(folds.test_start, expected_test_start)
        np.testing.assert_equal(folds.test_end, expected_test_end)


class TestFolds:
    def test_as_time(self):
        folds = Folds(
            timesteps=[
                datetime(2019, 1, 1),
                datetime(2019, 1, 2),
                datetime(2019, 1, 3),
                datetime(2019, 1, 4),
                datetime(2019, 1, 7),
                datetime(2019, 1, 8),
                datetime(2019, 1, 9),
                datetime(2019, 1, 10),
                datetime(2019, 1, 11),
                datetime(2019, 1, 14),
                datetime(2019, 1, 15),
            ],
            train_start=[0, 3],
            train_end=[5, 8],
            test_start=[6, 9],
            test_end=[7, 10],
        )
        actual = folds.as_time()
        np.testing.assert_equal(
            actual=actual.train_start,
            desired=[datetime(2019, 1, 1), datetime(2019, 1, 4)],
        )
        np.testing.assert_equal(
            actual=actual.train_end,
            desired=[datetime(2019, 1, 8), datetime(2019, 1, 11)],
        )
        np.testing.assert_equal(
            actual=actual.test_start,
            desired=[datetime(2019, 1, 9), datetime(2019, 1, 14)],
        )
        np.testing.assert_equal(
            actual=actual.test_end,
            desired=[datetime(2019, 1, 10), datetime(2019, 1, 15)],
        )



from tradingenv.contracts import Rate, ETF
from tradingenv.events import EventNBBO
from datetime import datetime
import pytest


class TestEventNBBO:
    def test_rate_does_not_raise_if_lower_than_25_pct(self):
        EventNBBO(datetime(2019, 1, 1), Rate('EONIA'), 0.24, 0.24)

    def test_rate_raise_if_greater_than_25_pct(self):
        with pytest.raises(ValueError):
            EventNBBO(datetime(2019, 1, 1), Rate('EONIA'), 0.25, 0.25)

    def test_raise_if_non_rate_is_negative(self):
        with pytest.raises(ValueError):
            EventNBBO(datetime(2019, 1, 1), ETF('SPY'), -1, -1)



from tradingenv.broker.fees import IBrokerFees
from tradingenv.broker.trade import Trade
from tradingenv.contracts import Cash, ETF, ES, ZN
from datetime import datetime, timedelta
import numpy as np
import pytest


class TestTrade:
    def test_initialization(self):
        now = datetime.now()
        spy = ETF("SPY")
        trade = Trade(now, spy, 1.0, 2.0, 3.0)
        assert trade.time == now
        assert trade.contract == spy
        assert trade.quantity == 1.0
        assert trade.bid_price == 2.0
        assert trade.ask_price == 3.0
        assert trade.notional == trade.quantity * spy.multiplier * trade.ask_price

    def test_notional_with_futures(self):
        es = ES(2019, 6)
        trade = Trade(datetime.now(), es, 1.0, 2.0, 3.0)
        assert trade.notional == trade.quantity * es.multiplier * trade.ask_price

    def test_raises_if_bid_price_is_nan(self):
        now = datetime.now()
        with pytest.raises(ValueError):
            Trade(now, ETF("SPY"), 1.0, np.nan, 3.0)

    def test_raises_if_ask_price_is_nan(self):
        now = datetime.now()
        with pytest.raises(ValueError):
            Trade(now, ETF("SPY"), 1.0, 2.0, np.nan)

    def test_raises_if_quantity_is_nan(self):
        now = datetime.now()
        with pytest.raises(ValueError):
            Trade(now, ETF("SPY"), np.nan, 2.0, 3.0)

    def test_raises_if_quantity_is_zero(self):
        now = datetime.now()
        with pytest.raises(ValueError):
            Trade(now, ETF("SPY"), 0.0, 2.0, 3.0)

    def test_raises_if_contract_is_cash(self):
        now = datetime.now()
        with pytest.raises(ValueError):
            Trade(now, Cash(), 0.0, 2.0, 3.0)

    def test_avg_price_when_buy(self):
        now = datetime.now()
        trade = Trade(now, ETF("SPY"), 1.0, 2.0, 3.0)
        assert trade.acq_price == 3.0

    def test_avg_price_when_sell(self):
        now = datetime.now()
        trade = Trade(now, ETF("SPY"), -1.0, 2.0, 3.0)
        assert trade.acq_price == 2.0

    def test_cost_of_cash_with_etf(self):
        spy = ETF("SPY")
        now = datetime.now()
        trade = Trade(now, spy, -1.5, 2.0, 3.0)
        assert trade.cost_of_cash == -1.5 * 2 * spy.multiplier * spy.cash_requirement

    def test_cost_of_cash_with_future(self):
        es = ES(2019, 6)
        now = datetime.now()
        trade = Trade(now, es, -1.5, 2.0, 3.0)
        assert trade.cost_of_cash == 0.0

    def test_cost_of_commissions_are_zero_by_default(self):
        now = datetime.now()
        trade = Trade(now, ETF("SPY"), 1.0, 2.0, 3.0)
        assert trade.cost_of_commissions == 0.0

    def test_cost_of_commissions_called_with_trade(self, mocker):
        class Fees(IBrokerFees):
            def commissions(self, trade: "Trade") -> float:
                return 0.42

        fees = Fees()
        mocker.patch.object(fees, "commissions")
        now = datetime.now()
        trade = Trade(now, ETF("SPY"), -5.0, 2.0, 4.0, fees)
        fees.commissions.called_once_with(trade)

    def test_cost_of_commissions(self):
        class Commissions(IBrokerFees):
            def commissions(self, trade: "Trade") -> float:
                return 0.42

        commissions = Commissions()
        now = datetime.now()
        trade = Trade(now, ETF("SPY"), -5.0, 2.0, 4.0, commissions)
        assert trade.cost_of_commissions == 0.42

    def test_cost_of_spread_when_negative_quantity(self):
        es = ES(2019, 6)
        now = datetime.now()
        trade = Trade(now, es, -1.5, 2.0, 3.0)
        actual = trade.cost_of_spread
        expected = (
            abs(trade.quantity)
            * es.multiplier
            * (trade.ask_price - trade.bid_price)
        )
        assert actual == expected

    def test_cost_of_spread_when_positive_quantity(self):
        es = ES(2019, 6)
        now = datetime.now()
        trade = Trade(now, es, 3, 2.0, 3.0)
        actual = trade.cost_of_spread
        expected = (
            trade.quantity * es.multiplier * (trade.ask_price - trade.bid_price)
        )
        assert actual == expected

    def test_equality(self):
        es = ES(2019, 6)
        now = datetime.now()
        assert Trade(now, es, 3, 2, 3) == Trade(now, es, 3, 2, 3)

    def test_equality_different_time(self):
        es = ES(2019, 6)
        now = datetime.now()
        assert Trade(now, es, 3, 2, 3) != Trade(now + timedelta(seconds=1), es, 3, 2, 3)

    def test_equality_different_contract(self):
        es = ES(2019, 6)
        zn = ZN(2019, 6)
        now = datetime.now()
        assert Trade(now, es, 3, 2, 3) != Trade(now, zn, 3, 2, 3)

    def test_equality_different_quantity(self):
        es = ES(2019, 6)
        now = datetime.now()
        assert Trade(now, es, 3, 2, 3) != Trade(now, es, 4, 2, 3)

    def test_equality_different_bid(self):
        es = ES(2019, 6)
        now = datetime.now()
        assert Trade(now, es, -3, 2, 3) != Trade(now, es, -3, 2.5, 3)

    def test_equality_different_ask(self):
        es = ES(2019, 6)
        now = datetime.now()
        assert Trade(now, es, 3, 2, 3) != Trade(now, es, 3, 2, 3.5)



"""The core module of trading-gym. TradingEnv is your market simulator to run
backtests or train reinforcement learning agents."""
from tradingenv.spaces import PortfolioSpace, BoxPortfolio
from tradingenv.broker.broker import Broker, EndOfEpisodeError
from tradingenv.broker.track_record import TrackRecord
from tradingenv.broker.fees import IBrokerFees, BrokerFees
from tradingenv.exchange import Exchange
from tradingenv.contracts import AbstractContract, Cash, Asset, Rate
from tradingenv.rewards import make_reward, AbstractReward, RewardSimpleReturn, LogReturn
from tradingenv.policy import make_policy
from tradingenv.state import IState, State
from tradingenv.features import Feature
from tradingenv.events import (
    Observer,
    IEvent,
    EventReset,
    EventStep,
    EventDone,
    EventNBBO,
    EventNewDate,
    EventNewObservation,
)
from tradingenv.transmitter import (
    AbstractTransmitter,
    AsynchronousTransmitter,
    Transmitter,
    TRAINING_SET,
)
from sklearn.preprocessing import PowerTransformer, StandardScaler
from sklearn.base import TransformerMixin
from sklearn.utils.validation import check_is_fitted
from sklearn.exceptions import NotFittedError
from typing import Tuple, Sequence, Union, List, Any, Optional
from pandas.core.generic import NDFrame
from datetime import datetime, timedelta
from tqdm import tqdm
from collections import deque, defaultdict
import pandas_market_calendars
import pandas as pd
import numpy as np
import threading
import gym


class TradingEnv(gym.Env):
    """TradingEnv is the core class of trading-gym, representing a market
    simulator to run backtests or train reinforcement learning agents."""

    metadata = {"render.modes": ["dashboard"]}

    def __init__(
        self,
        action_space: Union[List[AbstractContract], PortfolioSpace],
        state: Union[IState, List[Feature]] = IState(),
        reward: Union[str, AbstractReward] = RewardSimpleReturn(),
        transmitter: AbstractTransmitter = AsynchronousTransmitter(),
        prices: pd.DataFrame = None,
        initial_cash: float = 100,
        broker_fees: IBrokerFees = BrokerFees(),
        latency: float = 0,
        steps_delay: int = 0,
        fit_transformers: Union[bool, dict] = False,
        episode_length: int = None,
        sampling_span: int = None
    ):
        """
        Parameters
        ----------
        action_space: Union[List[AbstractContract], PortfolioSpace]
            This determines the investment universe of assets that can be
            traded during the simulations, as well any extra logic necessary to
            map an action to a portfolio of assets. To define a continuous
            long-only and unleveraged trading environment, it sufficient to
            pass a list of contracts to be traded. Supported contracts include:
            - tradingenv.contracts.Stock
            - tradingenv.contracts.ETF
            - tradingenv.contracts.Index
            - tradingenv.contracts.Future
            - tradingenv.contracts.FutureChain
            - tradingenv.contracts.AbstractContract for extra customisation.
            If you wish to allow short-selling or leveraged positions, you
            should pass one of the following instances:
            - tradingenv.spaces.DiscretePortfolio for discrete spaces
            - tradingenv.spaces.BoxPortfolio for continuous spaces
            - tradingenv.spaces.PortfolioSpace for extra customisation
            NOTE: there must be market data available for each contract being
            traded. Market data can be passed using the argument `prices` or
            `transmitter`.
        state: Union[IState, list[Feature]]
            This determines the observation space, which is empty by default.
            For backtesting purposes you don't need to specify this argument.
            For reinforcement learning training purposes, you can pass an
            tradingenv.state.IState object or a list of
            tradingenv.features.Feature.
        reward : Union[str, AbstractReward]
            The reward function to be used for reinforcement learning training.
            For backtesting purposes you don't need to specify this argument.
            Supported reward functions include:
            - tradingenv.rewards.RewardPnL
            - tradingenv.rewards.RewardLogReturn
            - tradingenv.rewards.RewardSimpleReturn
            - tradingenv.rewards.RewardDifferentialSharpeRatio
            - tradingenv.rewards.AbstractReward for extra customisation.
        transmitter : Transmitter
            This objects takes care of transmitting data and events at the right
            time during the simulations, as well as setting the timesteps
            at which the portfolio can be rebalanced. Use this argument to pass
            market prices, macroeconomic data or alternative data.
        prices : pd.DataFrame
            If your simulations only require market data, you can specify this
            argument instead of `transmitter`. Market prices can be passed as
            a pandas.DataFrame whose columns are contracts, index are
            timestamps and values are prices. However, this option assumes
            that bid-ask spread is zero. You are therefore recommended to use
            transmitter for more realistic simulations.
        initial_cash : float
            Each simulation starts with this amount of cash, 100 by default.
        broker_fees: IBrokerFees
            The simulation assumes absence of transaction costs by default. To
            set transaction costs including broker commissions, broker markup or
            a rate paid on idle cash you can pass an instance of
            tradingenv.broker.fees.BrokerFees.
        latency : float
            It defines the delay in seconds between the timestep at which the
            portfolio rebalancing decision was made and the time at which the
            trades to rebalance the portfolio are executed. This option can
            be useful when dealing with intraday data. For daily or even slower
            data you should use `steps_delay` instead.
        steps_delay : int
            Unlike `latency` which is expressed in seconds, this is expressed
            in amount of timesteps. For example, if a daily rebalancing
            frequency has been specified in the transmitter, then a rebalancing
            established in time t will be executed in t+1. This is set to
            zero by default to avoid surprised, but it is recommended to set
            this to 1 for more conservative and realistic simulations.
        fit_transformers : bool
            False by default. If True, observations defined by the `state`
            argument are using a class form sklearn.preprocessing.
        episode_length
            If not provided, the episode will terminate when the net liquidation
            value of the assets goes to zero or when market data stops. You
            can optionally provide a number of interaction with the environment
            after which the episode will terminate.
        sampling_span: int
            This argument is used only if episode_length is passed in
            TrandingEnv.reset. If specified, the episode start date is sampled
            using an exponentially decaying probability. sampling_span is the
            number of recent timesteps that captures ~70% of the likelihood of
            sampling a start date from such window. Uniform distribution is
            used by default if this parameter is not provided. It's useful to
            specify a value if you wish to train reinforcement learning agents
            that overweight observations from the more recent past.
        """
        if episode_length:
            # Adding 1 because there are (episode_length - 1) actions otherwise.
            episode_length += 1
        if not isinstance(action_space, PortfolioSpace):
            # action observation_space is assumed to be a sequence of contracts.
            action_space = BoxPortfolio(action_space)
        if prices is not None:
            prices.validate()  # source code in metrics.py
            transmitter = Transmitter(prices.index)
            transmitter.add_prices(prices)

        # Inputs.
        self.action_space: PortfolioSpace = action_space
        self.state = IState(state) if isinstance(state, list) else state
        self.observation_space = self.state.space
        self._verify_state = True
        self._reward = make_reward(reward)
        self._initial_cash = initial_cash
        self._broker_fees = broker_fees
        self._transmitter = transmitter
        self._latency = latency
        self._real_time = isinstance(transmitter, AsynchronousTransmitter)
        self._steps_delay =  steps_delay
        self._queue_actions: Optional[deque] = None
        self._sampling_span = sampling_span
        self._visits = defaultdict(int)
        self._episode_length = episode_length

        # Add extra events to transmitter and create partitions of events.
        for contract in self.action_space.contracts:
            self._transmitter.add_events(contract.make_events())
        self._transmitter._create_partitions(latency)

        # TODO: to improve UX, raise and error if there is a callback method
        #  registered to an event which does not exist anywhere in the
        #  transmitter.

        # Set by TradingEnv.reset.
        self._done: Union[bool, None] = None
        self.exchange: Union[Exchange, None] = None
        self.broker: Union[Broker, None] = None
        self._last_event: Union[IEvent, None] = None
        self._observers: Union[Sequence[Observer], None] = None
        self._now: Union[datetime, None] = None
        self._events_nonlatent: Union[List[IEvent], None] = None

        if fit_transformers:
            # Run procedure to fit transformers.
            # TODO: Test.
            kwargs = fit_transformers if isinstance(fit_transformers, dict) else dict()
            # needed to collect historical values of features
            self.backtest(**kwargs)
            for feature in self.state.features:
                feature.fit_transformer()
            # if transformers have been fit, this verification will fail.
            # Solution is to verify in parse by each feature and not here if
            # transformers have been fit.
            # By default, features before not transformation are verified anyway
            # even if verify state is False, so we should be still safe.
            self._verify_state = False

    def now(self) -> datetime:
        """Returns current time."""
        return self._transmitter._now() if self._real_time else self._now

    def reset(
            self,
            fold: str = TRAINING_SET,
            episode_length: int = None,
    ) -> IState:
        """Reset the environment to start a new episode. It's responsibility
        of the user to call this method before starting a new episode.

        Parameters
        ----------
        fold : str
            Only data from the specified fold will be used to simulate the
            episode. To get the full list of available folds check
            TradingEnv._transmitter.folds. By default, all data belong to a
            single fold named 'training-set'. Custom folds can be provided
            by instancing Transmitter with the optional argument 'folds'. You
            probably don't need to use this parameter unless you want to
            separate training from test data or run a walk-forward test.
        episode_length : int
            If specified, the episode will stop after this number of states.
            If not provided, the episode will last until when there are no
            more data (e.g. a complete pass trough the 'training-set' fold).

        Returns
        -------
        Initial state of the environment.
        """
        episode_length = episode_length or self._episode_length

        # Reset attributes.
        self._done = False
        self._last_event = None
        self._queue_actions = deque(
            [self.action_space.null_action() for _ in range(self._steps_delay)],
            maxlen=self._steps_delay + 1,
        )
        self._reward.reset()
        self.exchange = Exchange()
        self.broker = Broker(
            exchange=self.exchange,
            base_currency=self.action_space.base_currency,
            deposit=self._initial_cash,
            fees=self._broker_fees,
        )
        self.state.reset(self.exchange, self.action_space, self.broker)
        self._observers = (self.state, self.exchange)
        if self.state.features is not None:
            self._observers += tuple(self.state.features)
        self._now = None
        self._events_nonlatent = None

        # Process reset events.
        self.exchange.process_EventNBBO(EventNBBO(self.now(), Cash(), 1.0, 1.0))
        self.exchange.process_EventNBBO(EventNBBO(self.now(), self._broker_fees.interest_rate, 0.0, 0.0))

        # Instance generator of events to be sent at every interaction.
        self._transmitter._reset(fold, episode_length, self._sampling_span)
        self._events_latent, self._events_nonlatent = self._transmitter._next()
        self._process_latent_events()
        self._process_nonlatent_events()

        # Tear down events to notify observers.
        self.notify(EventReset(self.now(), self.broker.track_record))
        if self._done:
            self.notify(EventDone(self.now(), self.broker))
        return self.state(self._verify_state)

    def step(self, action: Union[int, np.ndarray]) -> Tuple[Any, float, bool, dict]:
        """
        Parameters
        ----------
        action : Union[int, np.ndarray]
            If the action observation_space is discrete, action is an integer
            representing the action ID. If the action observation_space is
            continuous, action is a np.array with the target weights of the
            portfolio to be rebalanced as indicated in TradingEnv.action_space.
            The weight will be applied to the notional trading value of the
            contract (transaction_price * multiplier).

        Returns
        -------
        A 4-tuple with: observation, reward, done, info.
        """
        if self._done:
            raise EndOfEpisodeError(
                "The current episode has ended. To start a new episode use "
                "TradingEnv.reset()."
            )
        self._queue_actions.appendleft(action)
        action = self._queue_actions.pop()
        self._process_latent_events()
        rebalancing = self.action_space.make_rebalancing_request(action, self.now(), self.broker)
        try:
            self.broker.rebalance(rebalancing)
        except EndOfEpisodeError:
            info = dict()
            self._done = True
        else:
            info = {"_rebalancing": rebalancing}
        self._process_nonlatent_events()
        reward = self._reward.calculate(self)

        # Tear down events to notify observers.
        self._visits[self.now()] += 1
        self.notify(EventStep(self.now(), self.broker.track_record, action))
        if self._done:
            # _process again to update the state with last data points.
            # Useful in a production environment, redundant otherwise.
            self.notify(EventDone(self.now(), self.broker))
        return self.state(self._verify_state), reward, self._done, info

    def render(
            self,
            mode: str = "dashboard",
            track_records: List[Union[str, TrackRecord]]=None,
            debug: bool=False,
            port: int=8050,
    ):
        """You have two options to visualize your results: (1) Use methods of
        TradingEnv.broker.track_record or (2) use TradingEnv.render()"""
        if mode not in self.metadata["render.modes"]:
            raise ValueError(
                "Mode {} is not supported. Supported modes are {}."
                "".format(mode, self.metadata["render.modes"])
            )
        if mode == 'dashboard':
            raise NotImplementedError("Legacy method that used to render "
                                      "results of the backtest or simulation.")
            # Note: there is no clean way for the user to stop this thread :(
            track_records = track_records or [self.broker.track_record]
            dashboard = Dashboard(track_records)
            thread = threading.Thread(
                daemon=True,
                target=dashboard.app.run_server,
                kwargs=dict(
                    debug=debug,
                    port=port,
                    dev_tools_silence_routes_logging=True,
                ),
            )
            thread.start()
            dashboard._thread = thread
            return dashboard

    def backtest(
        self,
        fold: str = TRAINING_SET,
        policy=None,
        episode_length: int = None,
        benchmark: NDFrame = None,
        risk_free: NDFrame = None,
        progress_bar: bool = False,
    ) -> "TrackRecord":
        """
        Run a backtest for a given policy. This method is useful for
        backtesting uses cases and not for reinforcement learning.

        Parameters
        ----------
        fold : str
            Only data from the specified fold will be used to simulate the
            episode. To get the full list of available folds check
            TradingEnv._transmitter.folds. By default, all data belong to a
            single fold named ''training-set'. Custom folds can be provided
            by instancing Transmitter with the optional argument 'folds'. You
            probably don't need to use this parameter unless you want to
            separate training from test data or run a walk-forward test.
        policy : TFPolicyGraph
            A an optional policy trained using the 'ray' framework. This is
            usually retrieved with agent.get_policy(). If not provided, actions
            will be randomly sampled from the action observation_space.
        episode_length : int
            If specified, the episode will stop after this number of states.
            If not provided, the episode will last until when there are no
            more data (e.g. a complete pass trough the 'training-set' fold).
        benchmark : NDFrame
            A NDFrame representing the levels series that you want to use to
            benchmark your investment strategy. This argument is mostly used
            when rendering the environment or calculating tearsheets.
        risk_free : NDFrame
            Risk free rate of the market (e.g. T-Bills, EURIBOR) expressed as
            a level pandas Series or DataFrame.
        progress_bar
            A boolean indicating whether or not a progress bar should be
            displayed while running the backtest. False by default.

        Returns
        -------
        An Episode instance. Interactions (state, action, reward) are stored
        using Barto and Sutton's notation: (S_t, A_t, R_t) where A_t has been
        computed from S_t and R_t is associated with A_t-1.
        """
        policy = make_policy(policy, self.action_space, self.observation_space)
        state = self.reset(fold, episode_length)
        done = False
        progress_bar_context = tqdm(
            total=len(self._transmitter._steps),
            disable=not progress_bar,
        )
        with progress_bar_context as progress_bar:
            progress_bar.update(1)
            while not done:
                now = self.now()
                progress_bar.set_description(
                    "[{start} : {end}] {policy}\t{now}"
                    "".format(
                        policy=policy,
                        start=self._transmitter._steps[0].date(),
                        now=now,
                        end=self._transmitter._steps[-1].date(),
                    )
                )
                action = policy.act(state)
                state, reward, done, info = self.step(action)
                progress_bar.update(1)
            progress_bar.update(1)

        # Returns track_record with metadata.
        # TODO: refactor.
        # TODO: Config stores hyperparams of the env such as BrokerFees,
        #  risk_free, traded contracts, action_space, ...
        track_record = self.broker.track_record
        track_record.fold = fold
        track_record.risk_free = risk_free
        track_record.benchmark = benchmark
        track_record.name = repr(policy)
        track_record.state_history = self.state.history
        return track_record

    def _is_new_date(self, time: datetime):
        # TODO: test/refactor.
        return (
            # True when resetting the environment, first event.
            self._last_event is not None and
            # In case time was not set during concrete event initialization.
            self._last_event.time is not None and
            time is not None and
            self._last_event.time is not np.nan and
            time is not np.nan and
            # Is new date.
            self._last_event.time.date() != time.date()
        )

    def notify(self, event: IEvent):
        """Add event to the queue to be processed by the dedicated thread.

        Parameters
        ----------
        event : IEvent
            An event to be processed by the environment. All features which are
            registered to observe this event (i.e. Feature instance implements
            the method process_<EventName>) will update their values
            accordingly.
        """
        AbstractContract.now = event.time
        self._now = event.time
        if self._is_new_date(event.time):
            # 'event' is the first event of the day. Notify that it's a new date
            self.notify(EventNewDate(self._last_event.time, self.broker))
        event.notify(self._observers)
        self._last_event = event

    def _process_latent_events(self):
        """These events are run before after the action of the agent AND before
        executing the rebalancing request in order to simulate a latency."""
        for event in self._events_latent:
            self.notify(event)
        self._events_latent = list()

    def _process_nonlatent_events(self):
        """Send events from the transmitter (if any). The episode ends if
        Transmitter has been iterated through all its timesteps. Continuing
        indefinitely if you are using AsynchronousTransmitter."""
        for event in self._events_nonlatent:
            self.notify(event)
        try:
            self._events_latent, self._events_nonlatent = self._transmitter._next()
        except StopIteration:
            self._done = True

    def visits(self) -> pd.Series:
        """Returns series indicating how many steps have been taken each
        day in the environment. """
        return pd.Series(self._visits).sort_index()


class TradingEnvXY(TradingEnv):
    """High-level wrapper of TradingEnv that allows to pass features in tabular
    form (X) and prices for assets to be traded (Y) as pandas.DataFrame."""

    def __init__(self,
                 X: pd.DataFrame,
                 Y: pd.DataFrame,
                 start: datetime = None,
                 end: datetime = None,
                 transformer: str = 'yeo-johnson',
                 transformer_end: str = None,
                 reward: str = 'logret',
                 cash: float = 100,
                 spread: float = 0.0002,
                 margin: float = 0.02,
                 markup: float = 0.005,
                 fee: float = 0.0002,
                 fixed: float = 0.,
                 rate: pd.Series = None,
                 latency: float = 0,
                 steps_delay: int = 1,
                 window: int = 1,
                 stride: int = None,
                 clip: float = 5.,
                 max_long: float = 1.,
                 max_short: float = -1.,
                 calendar: str = 'NYSE',
                 folds: dict = None,
                 episode_length: int = None,
                 sampling_span: int = None
                 ):
        """
        Note: we are loosely borrowing the notation of supervised learning
        where X is the input and Y is the output. While it remains true for X,
        in this application Y is a time series of prices to be traded.

        Parameters
        ----------
        X
            A pandas.DataFrame where index is time and columns are names of the
            exogenous variables. Missing values are allowed. The observation
            space of the environment will have as many variables as the number
            of columns.
        Y
            A pandas.Dataframe where index is time and columns are names of
            the assets being traded. Value are prices. Missing values are
            allowed.
        start
            The backtest of episodes cannot start earlier than this date.
        end
            The backtest of episodes cannot end after this date.
        transformer
            A sklearn.preprocessing transformer used to map the input data in
            a suitable range of deviation. As a shortcut, you can pass 'z-score'
            to standardise the data or 'yeo-johnson' for a power transform that
            will reduce the magnitude of outliers. Default is 'yeo-johnson'.
        transformer_end
            If a transformer is passed and is not fit, then the transformer
            will be fit over all available features data unitl transformer_end.
            If a value is not passed, then the transformer is fitted between
            over all available data until `end`.
        clip
            After having transformed the variables, values larger than this
            clip value will be truncated in a [-clip, +clip] range. The default
            clip value is 5.
        reward
            The reward function. At present, the only supported reawrd function
            is 'logreturn'.
        cash : float
            Each simulation starts with this amount of cash, 100 by default.
        spread
            Bid-ask spread of the assets being traded. It's 0.0002 by default,
            that is 0.02%.
        margin
            This is 0.02 by default, that is 2%. If the notional value of a
            trade over the net liquidation value of the trading account is
            smaller than the margin, then the trade will not be implemented.
            This can be useful to reduce transaction costs by filtering out
            small trades.
        rate
            A pandas series with the risk free rate used to calculate yield on
            idle cash or cost of leverage.
        markup
            Broker markup. If a risk free rate is provided, idle cash in the
            trading account will yield risk_free - markup. Viceversa, if the
            account is leveraged by borrowing from the broker, the cost of
            leverage will by risk_free + markup. This is 0.005 (0.5%) by
            default.
        fee
            The broker will charge a fee proportional to the notional traded
            value. This is 0.0002 (0.02%) by default.
        fixed : float
            Fixed broker fees to be applied to each trade, regardless the
            notional traded value of trade.
        latency : float
            It defines the delay in seconds between the timestep at which the
            portfolio rebalancing decision was made and the time at which the
            trades to rebalance the portfolio are executed. This option can
            be useful when dealing with intraday data. For daily or even slower
            data you should use `steps_delay` instead.
        steps_delay : int
            Number of days that it takes to implement trades. This is 1 by
            default. Note: if zero, then trades are placed in the same instant
            that information act upon becomes available.
        window
            The shape of the observation space will be (window, nr_features).
            Window is 1 by default. The observation will return the last
            `window` observation for each feature. Useful when past observation
            carry useful information.
        stride
            When the window size is greater than 1, the stride parameter is
            used to reduce the dimensionality of the observation. It determines
            the spacing between consecutive rows selected from the most recent
            time.
        max_long
            1 (100%) by default. Long a
            llocation in a given asset larger than
            max_long will raise an error.
        max_short
            -1 (-100%) by default. Short allocation in a given asset smaller
            than max_short will raise an error.
        calendar
            Trading calendar code. Prevents the backtest or interaction with
            the environment to occur when markets are closed. 'NYSE' by default.
        folds
            Used to partition the data in multiple folds.
        episode_length
            If not provided, the episode will terminate when the net liquidation
            value of the assets goes to zero or when market data stops. You
            can optionally provide a number of interaction with the environment
            after which the episode will terminate.
        sampling_span: int
            This argument is used only if episode_length is passed in
            TrandingEnv.reset. If specified, the episode start date is sampled
            using an exponentially decaying probability. sampling_span is the
            number of recent timesteps that captures ~70% of the likelihood of
            sampling a start date from such window. Uniform distribution is
            used by default if this parameter is not provided. It's useful to
            specify a value if you wish to train reinforcement learning agents
            that overweight observations from the more recent past.

        Attributes
        ----------
        X : pandas.DataFrame
            Data derived from the input X after having been transformed.
            The environment will use this data to calculate serve observations.
        Y : pandas.DataFrame
            Data derived from the input Y. The environment allows trading
            these asset prices after applying the bid-ask spread.
        """
        if isinstance(start, str):
            start = pd.to_datetime(start)
        start = start or Y.first_valid_index()
        start = max(start, Y.first_valid_index())
        if isinstance(end, str):
            end = pd.to_datetime(end)
        end = end or Y.last_valid_index()
        end = min(end, Y.last_valid_index())
        transformer_end = transformer_end or end
        if rate is None:
            rate = pd.Series(name='Zero Rate', dtype=float)
        rate = rate.squeeze().loc[start:end]
        rate.name = Rate(rate.name)
        if not rate.between(-1, +1).all():
            raise ValueError(
                "Argument `rate` is expressed as a percentage and it "
                "shouldn't. For example, 1% should be expressed as 0.01."
            )

        # Transform data.
        # TODO: Unit test no look-ahead bias in transformer.
        # TODO: If folds are provided, fit only on the training fold.
        if isinstance(transformer, TransformerMixin):
            self.transformer = transformer
        elif transformer == 'z-score':
            self.transformer = StandardScaler()
        elif transformer == 'yeo-johnson':
            self.transformer = PowerTransformer(transformer)
        else:
            raise ValueError(f"Unsupported transformer: {transformer}")
        self.transformer.set_output(transform="pandas")
        try:
            check_is_fitted(self.transformer)
        except NotFittedError:
            # Fit transformer using data before start but not after end
            # to provide more data to transformer and reduce 0-padding
            # when forward-filling low frequency data.
            self.transformer.fit(X.loc[:transformer_end])

        if reward == 'logret':
            scale = np.log(pd.DataFrame(Y).loc[start:transformer_end]).diff().std().mean().item()
            reward = LogReturn(scale=float(scale), clip=2.)
        else:
            raise NotImplementedError(f'Unsupported reward: {reward}')

        X = self.transformer.transform(X.loc[:end])
        X.ffill(inplace=True)
        X.fillna(0., inplace=True)
        X.clip(-clip, clip, inplace=True)

        # We drop features before the start date to speed-up warming-up
        # computations. Note if window is 1, then t0_warmup_idx is 0.
        t0 = X.loc[start:end].first_valid_index()
        t0_idx = X.index.get_loc(t0)
        t0_warmup_idx = t0_idx - window + 1
        assert t0_warmup_idx >= 0
        X = X.iloc[t0_warmup_idx:]

        # Discard data outside the data range and set contracts.
        Y = pd.DataFrame(Y).loc[start:end]
        Y.columns = [Asset(col) for col in Y.columns]
        assert X.first_valid_index() <= Y.first_valid_index()
        super().__init__(
            action_space=BoxPortfolio(Y.columns, max_short, max_long, margin=margin),
            state=State(X.columns.size, window, stride, max_=5),
            reward=reward,
            transmitter=self._make_transmitter(X, Y, calendar, spread, rate, folds, window),
            broker_fees=BrokerFees(markup, rate.name, fee, fixed),
            initial_cash=cash,
            latency=latency,
            steps_delay=steps_delay,
            episode_length=episode_length,
            sampling_span=sampling_span,
        )

        # Set attributes.
        self.X = X
        self.Y = Y
        self.start = min(self._transmitter.timesteps)
        self.end = max(self._transmitter.timesteps)

    @staticmethod
    def _make_timesteps(X: pd.DataFrame, Y: pd.DataFrame, calendar: str):
        # Drop dates where market is closed.
        calendar = pandas_market_calendars.get_calendar(calendar)
        holidays = calendar.holidays().holidays
        start = max(X.first_valid_index(), Y.first_valid_index())
        end = min(X.last_valid_index(), Y.last_valid_index())
        Y = Y.loc[start:end]
        timesteps = Y.drop([t for t in holidays if t in Y.index]).index
        return timesteps

    def _make_transmitter(self, X: pd.DataFrame, Y: pd.DataFrame, calendar: str, spread: float = 0., rate = None, folds = None, window = None):
        # TODO: test market calendar is applied.
        # TODO: test no past events are not processed - no need to spend computation in warming up if not needee.
        markov_reset = window == 1
        warmup = None if markov_reset else timedelta(days=3 + window * 2)
        timesteps = self._make_timesteps(X, Y, calendar)
        transmitter = Transmitter(timesteps, folds, markov_reset, warmup)
        for name in Y.columns:
            transmitter.add_prices(Y[[name]], spread)
        events = [EventNewObservation(t, x) for t, x in X.iterrows()]
        transmitter.add_events(events)
        transmitter.add_prices(rate.to_frame())
        return transmitter



"""By default, pytest runs only notebooks from \examples.
- \examples are lightweight to run (i.e. not computationally expensive)
and their main purpose is to show the trading-gym api.
- \applications are real-world application, which can be computationally
expensive to run.
"""


"""TradingEnv must receive events in order to update its state during a
simulation. This could either be done by the client with TradingEnv.add_event
(low-level) OR by providing a AbstractTransmitter when instancing TradingEnv (
high-level, recommended)."""
from tradingenv.events import IEvent, EventNBBO
from typing import Iterable, List, Dict, Type, Sequence, Tuple
from datetime import datetime, timezone
from collections import defaultdict, OrderedDict
from abc import ABC, abstractmethod
from datetime import timedelta
import numpy as np
import pandas as pd
import itertools
import bisect
import inspect
import random

TRAINING_SET = "training-set"
TEST_SET = "test-set"


class AbstractTransmitter(ABC):
    """Implementations of this interface take care of delivering events to
    the trading environment during the interaction of the agent with the
    environment at pre-specified timesteps."""

    @abstractmethod
    def _reset(self, fold_name: str = TRAINING_SET) -> None:
        """Iterator protocol."""

    @abstractmethod
    def _next(self) -> Tuple[List[IEvent], List[IEvent]]:
        """Iterator protocol."""

    @abstractmethod
    def _create_partitions(self, latency: float=0.) -> None:
        """Create partitions of Events, sorted chronologically."""

    @abstractmethod
    def _now(self) -> datetime:
        """Returns current time."""

    @abstractmethod
    def add_events(self, events: List[IEvent]):
        """"""


class AsynchronousTransmitter(AbstractTransmitter):
    """This AbstractTransmitter does not schedule any event to the TradingEnv.
    Suitable for when deploying to production your code."""

    def __len__(self) -> float:
        """This iterator never ends (it keeps returning empty lists), so len
        returns np.inf."""
        return np.inf

    def __repr__(self) -> str:
        return "{}(events: 0, timesteps: inf)".format(self.__class__.__name__)

    def _reset(self, *args, **kwargs):
        """Iterator protocol."""

    def _next(self) -> Tuple[List[IEvent], List[IEvent]]:
        """Iterator protocol. Returns an empty list."""
        return list(), list()

    def _create_partitions(self, latency: float=0.):
        """No need to create partitions."""

    def _now(self) -> datetime:
        """Returns current UTC time as a datetime object with timezone."""
        return datetime.utcnow()#.replace(tzinfo=timezone.utc)

    def add_events(self, events: List[IEvent]):
        return list()


class PartitionTimeRanges:
    def __init__(self, folds: dict = None):
        if folds is None:
            folds = {TRAINING_SET: [datetime.min, datetime.max]}

        # Parse split to ordered dictionary sorted by start date.
        self.folds = OrderedDict(sorted(folds.items(), key=lambda x: x[1]))
        self.verify_start_before_end()
        # I'm allowing partitions to overlap as some use cases such as
        # walk-forward partitions need support for this.
        # self.verify_overlaps()

    def __getitem__(self, item: str):
        return self.folds[item]

    def verify_start_before_end(self):
        """Verify that start < end across all datasets."""
        for key, (start, end) in self.folds.items():
            if end < start:
                raise ValueError(
                    "Dataset '{}' end date {} must be more recent than the "
                    "start date {}.".format(key, start, end)
                )

    def verify_overlaps(self):
        """Verify absence of any overlap across all time windows."""
        for key1, (start1, end1) in self.folds.items():
            for key2, (start2, end2) in self.folds.items():
                if key1 != key2 and start1 <= start2 <= end1:
                    raise ValueError(
                        "Dataset '{}' ({}:{})overlaps with '{}' ({}:{})"
                        "".format(key1, start1, end1, key2, start2, end2)
                    )


class Folds:
    def __init__(
        self,
        timesteps: Sequence[datetime],
        train_start: Sequence,
        train_end: Sequence,
        test_start: Sequence,
        test_end: Sequence,
    ):
        self.timesteps = timesteps
        self.train_start = train_start
        self.train_end = train_end
        self.test_start = test_start
        self.test_end = test_end

    def as_time(self):
        """Returns Folds instance storing timesteps associated with the end
        and start of each fold."""
        timesteps = np.array(self.timesteps)
        return Folds(
            timesteps=self.timesteps,
            train_start=timesteps[self.train_start],
            train_end=timesteps[self.train_end],
            test_start=timesteps[self.test_start],
            test_end=timesteps[self.test_end],
        )


class Transmitter(AbstractTransmitter):
    """AbstractTransmitter takes care of sending Event(s) to TradingEnv while
    interacting with the environment at specified timesteps. You want to use
    this class to run a backtest in order to schedule at what time the Event(s)
    (e.g. market prices or alternative data) should trigger.

    Attributes
    ----------
    events : List[Event]
        A list accumulating events to be delivered to TradingEnv during the
        interaction.
    timesteps : List[IEvent]
        A list accumulating timesteps at which to deliver events to
        TradingEnv during the interaction.
    """

    def __init__(
            self,
            timesteps: Sequence,
            folds: Dict[str, Sequence[datetime]] = None,
            markov_reset: bool = False,
            warmup: timedelta = None,
    ):
        """
        Parameters
        ----------
        timesteps : Sequence
            A sequence of timestamps (e.g. datetime, date, timestamp). Starting
            from the oldest timestamp, Transmitter next will send events in
            batches stopping at the next timestamp (chronological order) until
            when the last timestep is reached.
        folds : Dict[str, Sequence[datetime]]
            A dictionary mapping fold _names (keys) to a pair (Sequence) of
            timesteps defining the start and end time of the fold. If not
            provided, all timesteps will belong to a single fold named
            'training-set'.
        markov_reset : bool
            If False (default), all events occurring before the date of the
            reset will be processed to allow calculating derived data from past
            information. If True, past events will not be processed. Setting
            this to True will speed up the time required to reset the
            environment.

        Examples
        --------
        >>> from tradingenv.contracts import Index
        >>> returns = pd.DataFrame(
        ...     data=np.random.normal(0, 0.001, (100, 2)),
        ...     columns=[Index('S&P 500'), Index('T-Bond')],
        ...     index=pd.date_range('2019-01-01', periods=100, freq='B'),
        ... )
        >>> prices = (1 + returns).cumprod()
        >>>
        >>> # All prices belong to the training-set.
        >>> transmitter = Transmitter(timesteps=prices.index)
        >>> transmitter.add_prices(prices)

        >>> # Partition prices between training and test set.
        >>> transmitter = Transmitter(
        ...     timesteps=prices.index,
        ...     folds={
        ...         'training-set': [datetime(2019, 1, 1), datetime(2019, 3, 1)],
        ...         'test-set': [datetime(2019, 3, 2), datetime(2019, 5, 1)],
        ...     },
        ... )
        >>> transmitter.add_prices(prices)
        >>> transmitter._create_partitions()
        >>> transmitter._reset(fold_name='test-set')

        >>> # Retrieve batches of events, then processed by TradingEnv.add_event
        >>> events_batch_0 = transmitter._next()
        >>> events_batch_1 = transmitter._next()
        """
        # TODO: add optional argument events.
        # TODO: make timesteps optional.
        # TODO: is it an issue in case of mixed pd.Timestamp and datetime?
        # Inputs.
        self.timesteps: List[datetime] = list(timesteps)
        self._folds = PartitionTimeRanges(folds)
        self._markov_reset = markov_reset
        self._warmup = warmup

        # Attributes.
        self.events: List[IEvent] = list()
        self._fold_name: str = None
        self._start_date: datetime = None
        self._end_date: datetime = None
        self._partition_nonlatent: Dict[datetime, List[IEvent]] = None

    def __len__(self) -> int:
        """Length of the iterator, corresponding to the number of timesteps."""
        return len(self._partition_nonlatent.keys())

    def __repr__(self) -> str:
        nr_timesteps = "?" if self._partition_nonlatent is None else len(self._partition_nonlatent.keys())
        return "{cls}(events: {nr_events}, timesteps: {nr_timesteps})".format(
            cls=self.__class__.__name__,
            nr_events=len(self.events),
            nr_timesteps=nr_timesteps,
        )

    def _reset(self, fold_name: str = TRAINING_SET, episode_length: int = None, sampling_span: int = None):
        """
        Parameters
        ----------
        fold_name : str
            Fold id must have been specified in Transmitter.folds, which is
            'training-set' by default.
        episode_length : int
            Transmitter.next will stop returning events episode_length is
            reached. By default, there is no maximum length.
            If you are training reinforcement learning agents, setting this
            parameter might be good practice when episodes are very long
            (continuing tasks) or you want you use a RNN policy as training
            from long sequences might be harder for back-propagation through
            time.
        """
        if self._partition_nonlatent is None:
            self._create_partitions()
        start_date, end_date = self._folds[fold_name]
        timesteps = set(self._partition_nonlatent) | set(self._partition_latent)
        steps = np.sort(list(timesteps))
        steps = steps[start_date <= steps]
        steps = steps[steps <= end_date]
        if episode_length is not None:
            start_dates = steps[: -(episode_length - 1)]
            if sampling_span is not None:
                gamma = 1 - (1 / sampling_span)
                p = gamma ** np.arange(len(start_dates))[::-1]
                p /= p.sum()
            else:
                p = None
            start_date_idx = np.random.choice(range(len(start_dates)), p=p)
            end_date_idx = start_date_idx + episode_length - 1
            start_date = steps[start_date_idx]
            end_date = steps[end_date_idx]
            steps = steps[start_date_idx : end_date_idx + 1]

        # Done. Save to attributes.
        self._step_nr = 0
        self._current_time = np.nan
        self._steps = steps
        self._start_date = start_date
        self._end_date = end_date
        self._fold_name = fold_name

    def _next(self) -> Tuple[List[IEvent], List[IEvent]]:
        """Latent events front run the action to simulate a latency. Then the
        action if reflected in the trading environment, then non-latent events
        are run, then the new state is returned."""
        try:
            self._current_time = self._steps[self._step_nr]
        except IndexError:
            raise StopIteration()
        self._step_nr += 1
        if (self._step_nr == 1) and (not self._markov_reset):
            origin = (self._current_time - self._warmup) if self._warmup else datetime.min
            #start_date, end_date = self._folds[self._fold_name]
            events_latent = [
                e for t, e in self._partition_latent.items()
                if origin <= t <= self._current_time
            ]
            events_latent = list(itertools.chain(*events_latent))
            events_nonlatent = [
                events for t, events in self._partition_nonlatent.items()
                if origin <= t <= self._current_time
            ]
            events_nonlatent = list(itertools.chain(*events_nonlatent))
        else:
            events_latent = self._partition_latent[self._current_time]
            events_nonlatent = self._partition_nonlatent[self._current_time]
        return events_latent, events_nonlatent

    def _now(self) -> datetime:
        """Returns current UTC time as a datetime object with timezone."""
        return self._current_time

    def add_events(self, events: List[IEvent]) -> None:
        """
        Parameters
        ----------
        events : Iterable[IEvent]
            Events to be send to the environment. The time at which the Event
            will trigger corresponds to the first timestep greater or equal
            to Event.time.
        """
        self.events.extend(events)

    def add_timesteps(self, timesteps: List[datetime]) -> None:
        """
        Parameters
        ----------
        timesteps : Iterable[datetime]
            Events will be sent at the intervals specified here.
        """
        self.timesteps.extend(timesteps)

    def add_prices(self, prices: pd.DataFrame, spread: float = 0) -> None:
        """
        Parameters
        ----------
        prices : pandas.DataFrame
            A pandas.DataFrame with the following characteristics:
            - index: datetime objects
            - columns: contract _names
            - values: mid prices.
            All mid prices will be mapped to EventNBBOs to be delivered while
            interacting with TradingEnv. No timesteps are added.
        spread : float
            Spread to be applied to mid_prices (zero by default). For example,
            if spread=0.001, then there will be a 0.1% bid-ask spread
            for all prices.
        """
        prices: pd.DataFrame = prices.astype(float)
        for contract in prices.columns:
            series: pd.Series = prices[contract]
            series.dropna(inplace=True)
            for time, price in series.items():
                half_spread = price * spread / 2
                event = EventNBBO(
                    time=time,
                    contract=contract,
                    bid_price=price - half_spread,
                    ask_price=price + half_spread,
                    bid_size=np.inf,
                    ask_size=np.inf,
                )
                self.events.append(event)

    def add_custom_events(
        self,
        data: pd.DataFrame,
        event_class: Type[IEvent],
        mapping: Dict[str, str] = None,
    ):
        """
        Parameters
        ----------
        data : pandas.DataFrame
            Every row of this pandas.DataFrame will be used to create
            instance of EventType passing columns-state pairs as kwargs. The
            index of the dataframe must be of type Datetime and will be used
            to set Event.time.
        event_class : Type[Event]
            Arguments in EventType.__init__ will be searched from the columns
            of the dataframe and set accordingly.
        mapping : Dict[str, str]
            A dictionary mapping df column _names to attribute _names of
            EventType. There is no need to provide a mapping if the columns
            _names are exactly the same as the attribute _names of EventType.

        Examples
        --------
        >>> from tradingenv.events import IEvent
        >>> from tradingenv.transmitter import Transmitter
        >>> from datetime import datetime
        >>>
        >>> class NewsSentiment(IEvent):
        ...     def __init__(self, headline: str, sentiment: float):
        ...         self.headline = headline
        ...         self.sentiment = sentiment
        >>>
        >>> dataset = pd.DataFrame(
        ...     data=[
        ...         ['This is a very good news!', +0.7],
        ...         ['A kind of bad stuff has happened', -0.5],
        ...         ['Neutral news', 0],
        ...     ],
        ...     columns=['headline', 'sentiment'],
        ...     index=[
        ...         datetime(2019, 1, 1),
        ...         datetime(2019, 1, 1),
        ...         datetime(2019, 1, 2),
        ... ])
        >>> transmitter = Transmitter(timesteps=dataset.index)
        >>> transmitter.add_custom_events(dataset, NewsSentiment)
        """
        if mapping is None:
            mapping = dict()
        if not isinstance(data.index, pd.DatetimeIndex):
            raise TypeError("data.index must be of type pandas.DatetimeIndex.")
        parameters = dict(inspect.signature(event_class.__init__).parameters)
        parameters.pop("self")
        for index, row in data.iterrows():
            # If EventType requires one of the following missing keys during
            # its initialization, we use the index as a workaround.
            kwargs = row.to_dict()
            for key in ["Date", "date", "Time", "time"]:
                if key not in row.index:
                    kwargs[key] = index

            # Change attribute id according to mapping.
            for col_name, attr_name in mapping.items():
                kwargs[attr_name] = kwargs.pop(col_name)

            # Keep only relevant keys.
            kwargs = {k: v for k, v in kwargs.items() if k in parameters}

            # Create custom event.
            event = event_class(**kwargs)
            event.time = index
            self.events.append(event)

    def _min_timesteps_diff(self) -> float:
        """Returns minimum time gap in seconds between all timesteps."""
        gaps = np.diff(self.timesteps)
        return min([gap.total_seconds() for gap in gaps], default=np.inf)

    def _create_partitions(self, latency: float=0):
        """Create batches of events with bisect according to the timesteps."""
        if len(self.timesteps) == 0:
            raise ValueError(
                "Transmitter delivers events to TradingEnv at "
                "timesteps indicated by Transmitter.timesteps, "
                "which is empty. Have you called "
                "Transmitter.add_timesteps at least once?"
            )
        self.timesteps = sorted(set(self.timesteps))
        if latency >= self._min_timesteps_diff():
            raise ValueError(
                "Parameter 'latency'={}s must be smaller than the minimum "
                "time gap between each timesteps, corresponding to {}s."
                "".format(latency, self._min_timesteps_diff())
            )
        self._partition_nonlatent = defaultdict(list)
        self._partition_latent = defaultdict(list)
        # TODO: Test sorted. It guarantees that all events associated with a
        #  timestamp are sorted. This avoids possible surprises when processing
        #  events by Feature.
        events = sorted(e for e in self.events if e.time <= self.timesteps[-1])
        if self._markov_reset:
            events = [e for e in events if e.time >= self.timesteps[0]]
        for event in sorted(events):
            # Events which occur after the last timestep are never run.
            if event.time <= self.timesteps[-1]:
                # IEvent is associated to first timestep that is >= event.time
                index = bisect.bisect_left(self.timesteps, event.time)
                index_previous = index - 1
                timestep = self.timesteps[index]
                timestep_previous = self.timesteps[index_previous] if index_previous >= 0 else datetime(1800, 1, 1)  # or pd.Timestamp.min?
                sec_since_timestep = (event.time - timestep_previous).total_seconds()
                if sec_since_timestep < 0:
                    raise ValueError(
                        'Unexpected sec_since_timestep={}'
                        ''.format(sec_since_timestep)
                    )
                if sec_since_timestep <= latency:
                    self._partition_latent[timestep].append(event)
                else:
                    self._partition_nonlatent[timestep].append(event)

    def walk_forward(
        self, train_size: int = None, test_size: int = None, sliding_window: bool = True
    ) -> Folds:
        """
        Parameters
        ----------
        train_size : int
            Number of observations in the training folds.
        test_size : int
            Number of observation in the validation folds.
        sliding_window : bool
            If True (default), training folds will be computed using a sliding
            window with the same number of observations. If False, an
            expanding window will be used instead, so each training fold will
            start from the first timestep which a monotonically increasing
            number of observations.

        Notes
        -----
        By design, the step size of the window is set to be 'test_size'.
        Reasons are:
        - if 'step'<'test_size': your validation set is not using all available
        information.
        - if 'step'>'test_size': there is the risk of look-ahead bias because
        the next training fold will be validated on a portion of previously
        seen validation data.

        Returns
        -------
        A Folds object.
        """
        count = np.arange(len(self.timesteps))
        train_start = count[: -train_size - test_size + 1 : test_size]
        return Folds(
            timesteps=self.timesteps,
            train_start=train_start * int(sliding_window),
            train_end=train_start + train_size - 1,
            test_start=train_start + train_size,
            test_end=train_start + train_size + test_size - 1,
        )



from tradingenv.broker.rebalancing import Rebalancing
from tradingenv.broker.allocation import Weights, NrContracts
from tradingenv.contracts import ETF
from collections import namedtuple
from datetime import datetime
_TestSubtraction = namedtuple('TestSubtraction', ['right', 'expected'])


class TestRebalancing:
    def test_default_initialization(self):
        rebalancing = Rebalancing()
        assert rebalancing.allocation == dict()
        assert isinstance(rebalancing.allocation, Weights)
        assert rebalancing.absolute is True
        assert rebalancing.fractional is True
        assert rebalancing.margin == 0.0
        assert (rebalancing.time - datetime.now()).total_seconds() < 5

    def test_custom_initialization(self):
        now = datetime.now()
        rebalancing = Rebalancing(
            [ETF("SPY")], [0.8], 'nr-contracts', False, False, 0.02, now
        )
        assert rebalancing.allocation == {ETF("SPY"): 0.8}
        assert isinstance(rebalancing.allocation, NrContracts)
        assert rebalancing.absolute is False
        assert rebalancing.fractional is False
        assert rebalancing.margin == 0.02
        assert rebalancing.time == now



from tradingenv.exchange import LimitOrderBook, Exchange
from tradingenv.events import Observer, EventNBBO, EventContractDiscontinued
from tradingenv.contracts import FutureChain, ES, Index
from datetime import datetime, timezone
from collections import defaultdict
import numpy as np


class TestLimitOrderBook:
    def test_lazy_initialization(self):
        lob = LimitOrderBook()
        assert np.isnan(lob.bid_price)
        assert np.isnan(lob.ask_price)
        assert np.isnan(lob.bid_size)
        assert np.isnan(lob.ask_size)
        assert lob.time is None

        # Properties.
        assert np.isnan(lob.mid_price)
        assert np.isnan(lob.spread)

    def test_verbose_initialization(self):
        now = datetime.utcnow().replace(tzinfo=timezone.utc)
        lob = LimitOrderBook(1, 2, 3, 4, now)
        assert lob.bid_price == 1
        assert lob.ask_price == 2
        assert lob.bid_size == 3
        assert lob.ask_size == 4
        assert lob.time == now

        # Properties.
        assert lob.mid_price == 1.5
        assert lob.spread == 1


class TestExchange:
    def test_is_subclass_of_Observer(self):
        assert issubclass(Exchange, Observer)

    def test_initialization(self):
        exchange = Exchange()
        assert isinstance(exchange._books, defaultdict)
        assert exchange._books.default_factory == LimitOrderBook
        assert exchange.last_update is None
        assert exchange._observed_events == {
            "EventNBBO": "process_EventNBBO",
            "EventContractDiscontinued": "process_EventContractDiscontinued",
        }

    def test_getitem(self):
        exchange = Exchange()
        assert isinstance(exchange[Index("S&P 500")], LimitOrderBook)

    def test_len(self):
        exchange = Exchange()
        book1 = exchange._books[Index("S&P 500")]
        book2 = exchange._books["Bonds"]
        assert len(exchange) == 2

    def test_get_bid_prices(self):
        exchange = Exchange()
        exchange._books[Index("S&P 500")].bid_price = 9
        exchange._books[Index("S&P 500")].ask_price = 11
        exchange._books["Bonds"].bid_price = 99
        exchange._books["Bonds"].ask_price = 101
        expected = np.array([9, 99])
        actual = exchange.bid_prices([Index("S&P 500"), "Bonds"])
        assert all(actual == expected)

    def test_get_ask_prices(self):
        exchange = Exchange()
        exchange._books[Index("S&P 500")].bid_price = 9
        exchange._books[Index("S&P 500")].ask_price = 11
        exchange._books["Bonds"].bid_price = 99
        exchange._books["Bonds"].ask_price = 101
        expected = np.array([11, 101])
        actual = exchange.ask_prices([Index("S&P 500"), "Bonds"])
        assert all(actual == expected)

    def test_get_mid_prices(self):
        exchange = Exchange()
        exchange._books[Index("S&P 500")].bid_price = 9
        exchange._books[Index("S&P 500")].ask_price = 11
        exchange._books["Bonds"].bid_price = 99
        exchange._books["Bonds"].ask_price = 101
        expected = np.array([10, 100])
        actual = exchange.mid_prices([Index("S&P 500"), "Bonds"])
        assert all(actual == expected)

    def test_get_spreads(self):
        exchange = Exchange()
        exchange._books[Index("S&P 500")].bid_price = 9
        exchange._books[Index("S&P 500")].ask_price = 11
        exchange._books["Bonds"].bid_price = 99
        exchange._books["Bonds"].ask_price = 102
        expected = np.array([2, 3])
        actual = exchange.spreads([Index("S&P 500"), "Bonds"])
        assert all(actual == expected)

    def test_get_history(self):
        exchange = Exchange()
        exchange.process_EventNBBO(
            EventNBBO(datetime(2019, 1, 1), Index("S&P 500"), 1, 2, 3, 4)
        )
        exchange.process_EventNBBO(
            EventNBBO(datetime(2019, 1, 2), Index("T-Notes"), 5, 6, 7, 8)
        )
        exchange.process_EventNBBO(
            EventNBBO(datetime(2019, 1, 3), Index("T-Notes"), 9, 10, 11, 12)
        )
        exchange.process_EventNBBO(
            EventNBBO(datetime(2019, 1, 4), Index("S&P 500"), 13, 14, 15, 16)
        )

        assert exchange[Index("S&P 500")].history["time"] == [
            datetime(2019, 1, 1),
            datetime(2019, 1, 4),
        ]
        assert exchange[Index("S&P 500")].history["bid_price"] == [1, 13]
        assert exchange[Index("S&P 500")].history["ask_price"] == [2, 14]
        assert exchange[Index("S&P 500")].history["mid_price"] == [1.5, 13.5]
        assert exchange[Index("S&P 500")].history["bid_size"] == [3, 15]
        assert exchange[Index("S&P 500")].history["ask_size"] == [4, 16]

        assert exchange[Index("T-Notes")].history["time"] == [
            datetime(2019, 1, 2),
            datetime(2019, 1, 3),
        ]
        assert exchange[Index("T-Notes")].history["bid_price"] == [5, 9]
        assert exchange[Index("T-Notes")].history["ask_price"] == [6, 10]
        assert exchange[Index("T-Notes")].history["mid_price"] == [5.5, 9.5]
        assert exchange[Index("T-Notes")].history["bid_size"] == [7, 11]
        assert exchange[Index("T-Notes")].history["ask_size"] == [8, 12]

    def test_process_NBBO(self):
        now = datetime.utcnow().replace(tzinfo=timezone.utc)
        exchange = Exchange()
        exchange.process_EventNBBO(
            event=EventNBBO(
                time=now,
                bid_price=9,
                ask_price=11,
                bid_size=100,
                ask_size=200,
                contract=Index("S&P 500"),
            )
        )
        assert exchange._books[Index("S&P 500")].bid_price == 9
        assert exchange._books[Index("S&P 500")].ask_price == 11
        assert exchange._books[Index("S&P 500")].bid_size == 100
        assert exchange._books[Index("S&P 500")].ask_size == 200
        assert exchange._books[Index("S&P 500")].time == now
        assert exchange.last_update == now

    def test_process_EventContractDiscontinued(self):
        exchange = Exchange()
        exchange.process_EventNBBO(
            event=EventNBBO(
                time=datetime.now(),
                contract=Index("S&P 500"),
                bid_price=9,
                ask_price=11,
                bid_size=100,
                ask_size=200,
            )
        )
        assert exchange[Index("S&P 500")].bid_price == 9
        history_before_termination = exchange[Index("S&P 500")].history
        exchange.process_EventContractDiscontinued(
            event=EventContractDiscontinued(datetime.now(), Index("S&P 500"))
        )
        history_after_termination = exchange[Index("S&P 500")].history
        assert exchange[Index("S&P 500")].bid_price is np.nan
        assert history_before_termination == history_after_termination

    def test_process_NBBO_doesnt_update_after_discontinuation_of_book(self):
        event_discontinued = EventContractDiscontinued(datetime.now(), Index("S&P 500"))
        event_nbbo = EventNBBO(
            time=datetime.now(),
            contract=Index("S&P 500"),
            bid_price=9,
            ask_price=11,
            bid_size=100,
            ask_size=200,
        )

        exchange = Exchange()

        exchange.process_EventNBBO(event_nbbo)
        assert exchange[Index("S&P 500")].bid_price == 9

        exchange.process_EventContractDiscontinued(event_discontinued)
        assert exchange[Index("S&P 500")].bid_price is np.nan

        exchange.process_EventNBBO(event_nbbo)
        assert exchange[Index("S&P 500")].bid_price is np.nan  # nan despite nbbo

    def test_getitem_changes_key_for_leading_contracts(self):
        lead_contract = FutureChain(ES, "2018-12", "2019-03")
        lead_contract.now = datetime.min
        exchange = Exchange()
        assert exchange[lead_contract] is exchange[ES(2018, 12)]
        assert exchange[lead_contract] is not exchange[ES(2019, 3)]



from tradingenv.transmitter import AbstractTransmitter, Transmitter
from tradingenv.contracts import Index
from tradingenv.events import IEvent, EventNBBO
from datetime import datetime, timedelta
import pandas as pd
import pytest
import numpy as np
import itertools


class MarketEvent(IEvent):
    """Concrete event with timestamp, used for testing."""

    def __init__(self, time: datetime):
        self.time = time


class TestTransmitter:
    def test_parent_class_is_Transmitter(self):
        assert issubclass(Transmitter, AbstractTransmitter)

    def test_add_events(self):
        event1 = MarketEvent(datetime(2019, 1, 1))
        event2 = MarketEvent(datetime(2019, 1, 2))
        event3 = MarketEvent(datetime(2019, 1, 2))
        transmitter = Transmitter([datetime(2019, 1, 1, 12), datetime(2019, 1, 2)])
        transmitter.add_events([event1, event2, event3])
        transmitter._create_partitions()
        transmitter._reset()
        assert transmitter._now() is np.nan
        assert transmitter._next() == ([], [event1])
        assert transmitter._now() == datetime(2019, 1, 1, 12)
        assert transmitter._next() == ([], [event2, event3])
        assert transmitter._now() == datetime(2019, 1, 2)
        with pytest.raises(StopIteration):
            transmitter._next()

    def test_markov_walk_forward_partition(self):
        event1 = MarketEvent(datetime(2019, 1, 1))
        event2 = MarketEvent(datetime(2019, 1, 2))
        event3 = MarketEvent(datetime(2019, 1, 3))
        transmitter = Transmitter(
            timesteps=[datetime(2019, 1, 1), datetime(2019, 1, 2), datetime(2019, 1, 3)],
            folds={
                "2019": [datetime(2019, 1, 1), datetime(2019, 1, 2)],
                "2020": [datetime(2019, 1, 2), datetime(2019, 1, 3)],
            },
            markov_reset=True,
        )
        transmitter.add_events([event1, event2, event3])
        transmitter._reset('2020')
        _, events = transmitter._next()
        assert events == [event2]  #and not [event1]

    @pytest.mark.parametrize("markov_reset,expected", [
        [False, [datetime(2019, 1, 1), datetime(2019, 1, 2)]],
        [True, [datetime(2019, 1, 2)]]
    ])
    def test_markov_reset(self, markov_reset, expected):
        event1 = MarketEvent(datetime(2019, 1, 1))
        event2 = MarketEvent(datetime(2019, 1, 2))
        event3 = MarketEvent(datetime(2019, 1, 3))
        timesteps = [datetime(2019, 1, 2), datetime(2019, 1, 3)]
        transmitter = Transmitter(timesteps, markov_reset=markov_reset)
        transmitter.add_events([event1, event2, event3])
        transmitter._reset()
        [], events = transmitter._next()
        actual = [event.time for event in events]
        assert actual == expected

    def test_create_partitions_ignores_events_after_last_timestep(self):
        event = MarketEvent(datetime(2019, 1, 2))
        transmitter = Transmitter([datetime(2019, 1, 1)])
        transmitter.add_events([event])
        transmitter._create_partitions()  # does not raise
        assert transmitter._partition_nonlatent == {}

    def test_duplicate_timesteps_are_dropped(self):
        event1 = MarketEvent(datetime(2019, 1, 1))
        event2 = MarketEvent(datetime(2019, 1, 2))
        event3 = MarketEvent(datetime(2019, 1, 2))
        transmitter = Transmitter(
            [
                datetime(2019, 1, 1, 12),
                datetime(2019, 1, 2),
                datetime(2019, 1, 2),  # DUPLICATE
            ]
        )
        transmitter.add_events([event1, event2, event3])
        transmitter._create_partitions()
        transmitter._reset()
        assert transmitter._now() is np.nan
        assert transmitter._next() == ([], [event1])
        assert transmitter._now() == datetime(2019, 1, 1, 12)
        assert transmitter._next() == ([], [event2, event3])
        assert transmitter._now() == datetime(2019, 1, 2)
        with pytest.raises(StopIteration):
            transmitter._next()

    def test_add_disordered_events(self):
        event1 = MarketEvent(datetime(2019, 1, 1))
        event2 = MarketEvent(datetime(2019, 1, 2))
        event3 = MarketEvent(datetime(2019, 1, 2))
        transmitter = Transmitter([datetime(2019, 1, 1, 12), datetime(2019, 1, 2)])
        transmitter.add_events([event2, event1, event3])
        transmitter._create_partitions()
        transmitter._reset()
        assert transmitter._now() is np.nan
        assert transmitter._next() == ([], [event1])
        assert transmitter._now() == datetime(2019, 1, 1, 12)
        assert transmitter._next() == ([], [event2, event3])
        assert transmitter._now() == datetime(2019, 1, 2)
        with pytest.raises(StopIteration):
            transmitter._next()

    def test_add_events_without_timesteps_raises_error(self):
        event1 = MarketEvent(datetime(2019, 1, 1))
        event2 = MarketEvent(datetime(2019, 1, 2))
        event3 = MarketEvent(datetime(2019, 1, 2))
        transmitter = Transmitter(timesteps=list())
        transmitter.add_events([event2, event1, event3])
        with pytest.raises(ValueError):
            transmitter._create_partitions()

    def test_add_disordered_timesteps_and_disordered_events(self):
        # Events are sorted by time, so let's sleep a bit to make sure that
        # their attribute 'time' is strictly increasing over time.
        event1 = MarketEvent(datetime(2019, 1, 1))
        event2 = MarketEvent(datetime(2019, 1, 2))
        event3 = MarketEvent(datetime(2019, 1, 2))
        transmitter = Transmitter([datetime(2019, 1, 2), datetime(2019, 1, 1, 12)])
        transmitter.add_events([event3, event1, event2])
        transmitter._create_partitions()
        transmitter._reset()
        assert transmitter._now() is np.nan
        assert transmitter._next() == ([], [event1])
        assert transmitter._now() == datetime(2019, 1, 1, 12)
        [], events = transmitter._next()
        assert (events == [event2, event3]) or (events == [event3, event2])
        assert transmitter._now() == datetime(2019, 1, 2)
        with pytest.raises(StopIteration):
            transmitter._next()

    def test_sequence_resets_now_after_iteration_end(self):
        event1 = MarketEvent(datetime(2019, 1, 1))
        event2 = MarketEvent(datetime(2019, 1, 2))
        event3 = MarketEvent(datetime(2019, 1, 2))
        transmitter = Transmitter([datetime(2019, 1, 1, 12), datetime(2019, 1, 2)])
        transmitter.add_events([event1, event2, event3])
        transmitter._create_partitions()
        for _ in range(10):
            transmitter._reset()
            assert transmitter._now() is np.nan
            assert transmitter._next() == ([], [event1])
            assert transmitter._now() == datetime(2019, 1, 1, 12)
            assert transmitter._next() == ([], [event2, event3])
            assert transmitter._now() == datetime(2019, 1, 2)
            with pytest.raises(StopIteration):
                transmitter._next()

    def test_add_prices(self):
        prices = pd.DataFrame(
            data=[[1, 2], [np.nan, 4], [5, "6"], [None, 8]],
            index=[
                datetime(2019, 1, 5),
                datetime(2019, 1, 7),
                datetime(2019, 1, 9),
                datetime(2019, 1, 11),
            ],
            columns=[Index("S&P 500"), Index("T-bills")],
        )
        transmitter = Transmitter([datetime(2019, 1, 7, 12), datetime(2019, 1, 11)])
        transmitter.add_prices(prices=prices, spread=0.01)
        transmitter._create_partitions()

        # Test.
        transmitter._reset()
        assert transmitter._now() is np.nan
        assert transmitter._next() == ([], [
            EventNBBO(datetime(2019, 1, 5), Index("S&P 500"), 0.995, 1.005, np.inf, np.inf),
            EventNBBO(datetime(2019, 1, 5), Index("T-bills"), 1.990, 2.010, np.inf, np.inf),
            EventNBBO(datetime(2019, 1, 7), Index("T-bills"), 3.980, 4.020, np.inf, np.inf),
        ])
        assert transmitter._now() == datetime(2019, 1, 7, 12)
        assert transmitter._next() == ([], [
            EventNBBO(datetime(2019, 1, 9), Index("S&P 500"), 4.975, 5.025, np.inf, np.inf),
            EventNBBO(datetime(2019, 1, 9), Index("T-bills"), 5.970, 6.030, np.inf, np.inf),
            EventNBBO(datetime(2019, 1, 11), Index("T-bills"), 7.960, 8.040, np.inf, np.inf),
        ])
        assert transmitter._now() == datetime(2019, 1, 11)
        with pytest.raises(StopIteration):
            transmitter._next()

    def test_add_custom_events_with_time_in_init_args(self):
        class NewsSentiment(IEvent):
            def __init__(self, time: datetime, headline: str, sentiment: float):
                self.time = time
                self.headline = headline
                self.sentiment = sentiment

            def __eq__(self, other: "NewsSentiment"):
                return all(
                    [
                        self.time == other.time,
                        self.headline == other.headline,
                        self.sentiment == other.sentiment,
                    ]
                )

        dataset = pd.DataFrame(
            data=[
                ["This is a very good news!", +0.7],
                ["A kind of bad stuff has happened", -0.5],
                ["Neutral news", 0],
            ],
            columns=["headline", "sentiment"],
            index=[datetime(2019, 1, 1), datetime(2019, 1, 1), datetime(2019, 1, 2)],
        )
        transmitter = Transmitter(timesteps=dataset.index)
        transmitter.add_custom_events(dataset, NewsSentiment)
        transmitter._create_partitions()
        transmitter._reset()

        pre_latency, actual = transmitter._next()
        assert pre_latency == list()
        expected = [
            NewsSentiment(datetime(2019, 1, 1), "This is a very good news!", 0.7),
            NewsSentiment(
                datetime(2019, 1, 1), "A kind of bad stuff has happened", -0.5
            ),
        ]
        assert actual == expected

        pre_latency, actual = transmitter._next()
        assert pre_latency == list()
        expected = [NewsSentiment(datetime(2019, 1, 2), "Neutral news", 0)]
        assert actual == expected
        with pytest.raises(StopIteration):
            transmitter._next()

    def test_add_custom_events_without_time_in_init_args(self):
        class NewsSentiment(IEvent):
            def __init__(self, headline: str, sentiment: float):
                self.headline = headline
                self.sentiment = sentiment

            def __eq__(self, other: "NewsSentiment"):
                return all(
                    [
                        self.time == other.time,
                        self.headline == other.headline,
                        self.sentiment == other.sentiment,
                    ]
                )

        dataset = pd.DataFrame(
            data=[
                ["This is a very good news!", +0.7],
                ["A kind of bad stuff has happened", -0.5],
                ["Neutral news", 0],
            ],
            columns=["headline", "sentiment"],
            index=[datetime(2019, 1, 1), datetime(2019, 1, 1), datetime(2019, 1, 2)],
        )
        transmitter = Transmitter(timesteps=dataset.index)
        transmitter.add_custom_events(dataset, NewsSentiment)
        transmitter._create_partitions()
        transmitter._reset()

        event1 = NewsSentiment("This is a very good news!", 0.7)
        event1.time = datetime(2019, 1, 1)
        event2 = NewsSentiment("A kind of bad stuff has happened", -0.5)
        event2.time = datetime(2019, 1, 1)
        expected = [event1, event2]
        pre_latency, actual = transmitter._next()
        assert pre_latency == list()
        assert actual == expected

        event = NewsSentiment("Neutral news", 0)
        event.time = datetime(2019, 1, 2)
        pre_latency, actual = transmitter._next()
        assert pre_latency == list()
        expected = [event]
        assert actual == expected

        with pytest.raises(StopIteration):
            transmitter._next()

    def test_add_custom_events_raises_if_index_is_not_datetime(self):
        class NewsSentiment(IEvent):
            def __init__(self, time: datetime, headline: str, sentiment: float):
                self.time = time
                self.headline = headline
                self.sentiment = sentiment

            def __eq__(self, other: "NewsSentiment"):
                return all(
                    [
                        self.time == other.time,
                        self.headline == other.headline,
                        self.sentiment == other.sentiment,
                    ]
                )

        dataset = pd.DataFrame(
            data=[
                ["This is a very good news!", +0.7],
                ["A kind of bad stuff has happened", -0.5],
                ["Neutral news", 0],
            ],
            columns=["headline", "sentiment"],
            index=[1, 2, 3],  # should be datetime, so TypeError is expected
        )
        transmitter = Transmitter(timesteps=dataset.index)
        with pytest.raises(TypeError):
            # TypeError('df.index must be of type pandas.DatetimeIndex.')
            transmitter.add_custom_events(dataset, NewsSentiment)

    def test_add_custom_events_uses_provided_mapping(self):
        class NewsSentiment(IEvent):
            def __init__(self, time: datetime, headline: str, sentiment: float):
                self.time = time
                self.headline = headline
                self.sentiment = sentiment

            def __eq__(self, other: "NewsSentiment"):
                return all(
                    [
                        self.time == other.time,
                        self.headline == other.headline,
                        self.sentiment == other.sentiment,
                    ]
                )

        dataset = pd.DataFrame(
            data=[
                ["This is a very good news!", +0.7],
                ["A kind of bad stuff has happened", -0.5],
                ["Neutral news", 0],
            ],
            columns=["col1", "col2"],
            index=[datetime(2019, 1, 1), datetime(2019, 1, 1), datetime(2019, 1, 2)],
        )
        transmitter = Transmitter(timesteps=dataset.index)
        transmitter.add_custom_events(
            dataset, NewsSentiment, mapping={"col1": "headline", "col2": "sentiment"}
        )
        transmitter._create_partitions()
        transmitter._reset()

        pre_latency, actual = transmitter._next()
        assert pre_latency == list()
        expected = [
            NewsSentiment(datetime(2019, 1, 1), "This is a very good news!", 0.7),
            NewsSentiment(
                datetime(2019, 1, 1), "A kind of bad stuff has happened", -0.5
            ),
        ]
        assert actual == expected

        pre_latency, actual = transmitter._next()
        assert pre_latency == list()
        expected = [NewsSentiment(datetime(2019, 1, 2), "Neutral news", 0)]
        assert actual == expected

    def test_split_date(self):
        prices = pd.DataFrame(
            data=[[1, 2], [3, 4], [5, 6], [10, 11], [20, 21], [22, 23]],
            index=[
                datetime(2019, 1, 1),
                datetime(2019, 1, 2),
                datetime(2019, 1, 3),
                datetime(2019, 1, 4),
                datetime(2019, 1, 5),
                datetime(2019, 1, 6),
            ],
            columns=[Index("S&P 500"), Index("T-bills")],
        )
        transmitter = Transmitter(
            timesteps=prices.index,
            folds={
                "training-set": (datetime(2019, 1, 1), datetime(2019, 1, 3)),
                "validation-set": (datetime(2019, 1, 4), datetime(2019, 1, 4)),
                "test-set": (datetime(2019, 1, 5), datetime(2019, 1, 6)),
            },
        )
        transmitter.add_prices(prices)
        transmitter._create_partitions()

        # Reset uses training data by default..
        transmitter._reset()
        assert transmitter._now() is np.nan

        assert transmitter._next() == ([], [
            EventNBBO(datetime(2019, 1, 1), Index("S&P 500"), 1, 1),
            EventNBBO(datetime(2019, 1, 1), Index("T-bills"), 2, 2),
        ])
        assert transmitter._now() == datetime(2019, 1, 1)

        assert transmitter._next() == ([], [
            EventNBBO(datetime(2019, 1, 2), Index("S&P 500"), 3, 3),
            EventNBBO(datetime(2019, 1, 2), Index("T-bills"), 4, 4),
        ])
        assert transmitter._now() == datetime(2019, 1, 2)

        assert transmitter._next() == ([], [
            EventNBBO(datetime(2019, 1, 3), Index("S&P 500"), 5, 5),
            EventNBBO(datetime(2019, 1, 3), Index("T-bills"), 6, 6),
        ])
        assert transmitter._now() == datetime(2019, 1, 3)

        with pytest.raises(StopIteration):
            transmitter._next()

        # Test data.
        transmitter._reset("test-set")
        assert transmitter._now() is np.nan

        assert transmitter._next() == ([], [
            EventNBBO(datetime(2019, 1, 1), Index("S&P 500"), 1, 1),
            EventNBBO(datetime(2019, 1, 1), Index("T-bills"), 2, 2),
            EventNBBO(datetime(2019, 1, 2), Index("S&P 500"), 3, 3),
            EventNBBO(datetime(2019, 1, 2), Index("T-bills"), 4, 4),
            EventNBBO(datetime(2019, 1, 3), Index("S&P 500"), 5, 5),
            EventNBBO(datetime(2019, 1, 3), Index("T-bills"), 6, 6),
            EventNBBO(datetime(2019, 1, 4), Index("S&P 500"), 10, 10),
            EventNBBO(datetime(2019, 1, 4), Index("T-bills"), 11, 11),
            EventNBBO(datetime(2019, 1, 5), Index("S&P 500"), 20, 20),
            EventNBBO(datetime(2019, 1, 5), Index("T-bills"), 21, 21),
        ])
        assert transmitter._now() == datetime(2019, 1, 5)

        assert transmitter._next() == ([], [
            EventNBBO(datetime(2019, 1, 6), Index("S&P 500"), 22, 22),
            EventNBBO(datetime(2019, 1, 6), Index("T-bills"), 23, 23),
        ])
        assert transmitter._now() == datetime(2019, 1, 6)

        with pytest.raises(StopIteration):
            transmitter._next()

        # Validation data.
        transmitter._reset("validation-set")
        assert transmitter._now() is np.nan

        assert transmitter._next() == ([], [
            EventNBBO(datetime(2019, 1, 1), Index("S&P 500"), 1, 1),
            EventNBBO(datetime(2019, 1, 1), Index("T-bills"), 2, 2),
            EventNBBO(datetime(2019, 1, 2), Index("S&P 500"), 3, 3),
            EventNBBO(datetime(2019, 1, 2), Index("T-bills"), 4, 4),
            EventNBBO(datetime(2019, 1, 3), Index("S&P 500"), 5, 5),
            EventNBBO(datetime(2019, 1, 3), Index("T-bills"), 6, 6),
            EventNBBO(datetime(2019, 1, 4), Index("S&P 500"), 10, 10),
            EventNBBO(datetime(2019, 1, 4), Index("T-bills"), 11, 11),
        ])
        assert transmitter._now() == datetime(2019, 1, 4)

        with pytest.raises(StopIteration):
            transmitter._next()

    @pytest.mark.parametrize(
        argnames=["fold_name", "valid_origins"],
        argvalues=[
            [
                "training-set",
                [datetime(2019, 1, 1), datetime(2019, 1, 2), datetime(2019, 1, 3)],
            ],
            [
                "test-set",
                [datetime(2019, 1, 5), datetime(2019, 1, 6), datetime(2019, 1, 7)],
            ],
        ],
    )
    def test_episode_length_as_int(self, fold_name, valid_origins):
        prices = pd.DataFrame(
            data=[
                [1, 2],
                [3, 4],
                [5, 6],
                [10, 11],
                [20, 21],
                [22, 23],
                [24, 25],
                [27, 29],
            ],
            index=[
                datetime(2019, 1, 1),
                datetime(2019, 1, 2),
                datetime(2019, 1, 3),
                datetime(2019, 1, 4),
                datetime(2019, 1, 5),
                datetime(2019, 1, 6),
                datetime(2019, 1, 7),
                datetime(2019, 1, 8),
            ],
            columns=[Index("S&P 500"), Index("T-bills")],
        )
        transmitter = Transmitter(
            timesteps=prices.index,
            folds={
                "training-set": [datetime(2019, 1, 1), datetime(2019, 1, 4)],
                "test-set": [datetime(2019, 1, 5), datetime(2019, 1, 8)],
            },
        )
        transmitter.add_prices(prices)
        transmitter._create_partitions()

        origin_list = list()
        for _ in range(10000):
            transmitter._reset(fold_name, episode_length=2)
            events_latent, events_nonlatent = transmitter._next()

            # Assert start date is within the fold.
            origin = transmitter._now()
            assert origin in valid_origins

            # Assert latent events include all events of the fold up to _now
            start_date, end_date = transmitter._folds[transmitter._fold_name]
            start_fold_to_current_times = [
                t for t in transmitter._partition_latent
                if t <= transmitter._current_time
            ]
            expected_events_latent = [transmitter._partition_latent[t] for t in start_fold_to_current_times]
            expected_events_latent = list(itertools.chain(*expected_events_latent))
            assert expected_events_latent == events_latent

            # Assert nonlatent events include all events of the fold up to _now
            start_date, end_date = transmitter._folds[transmitter._fold_name]
            start_fold_to_current_times = [
                t for t in transmitter._partition_nonlatent
                if t <= transmitter._current_time
            ]
            expected_events_nonlatent = [transmitter._partition_nonlatent[t] for t in start_fold_to_current_times]
            expected_events_nonlatent = list(itertools.chain(*expected_events_nonlatent))
            assert expected_events_nonlatent == events_nonlatent

            # Episode length is 2, so at next iteration is expected to stop.
            transmitter._next()
            assert transmitter._now() == origin + timedelta(days=1)
            with pytest.raises(StopIteration):
                transmitter._next()
            origin_list.append(origin)

        # All relative frequencies of origin must be the same.
        relative_freq = pd.Series(origin_list).value_counts(normalize=True)
        np.testing.assert_allclose(
            actual=relative_freq.values,
            desired=np.full(
                shape=(len(valid_origins),), fill_value=1 / len(valid_origins)
            ),
            rtol=0,
            atol=0.02,
        )

    def test_create_partitions_without_latency(self):
        transmitter = Transmitter(
            timesteps=[
                # Day 1.
                datetime(2019, 1, 1, 15, 59),
                datetime(2019, 1, 1, 16),
                # Day 2.
                datetime(2019, 1, 2, 15, 59),
                datetime(2019, 1, 2, 16),
            ])
        events = list()
        for t in transmitter.timesteps:
            for delay in [-1, 0, 1, 29, 30, 31]:
                events.append(MarketEvent(t+timedelta(seconds=delay)))
        transmitter.add_events(events)
        transmitter._create_partitions(latency=0)
        assert transmitter._partition_nonlatent == {
            datetime(2019, 1, 1, 15, 59): [
                MarketEvent(datetime(2019, 1, 1, 15, 58, 59)),
                MarketEvent(datetime(2019, 1, 1, 15, 59))],
            datetime(2019, 1, 1, 16, 0): [
                MarketEvent(datetime(2019, 1, 1, 15, 59, 1)),
                MarketEvent(datetime(2019, 1, 1, 15, 59, 29)),
                MarketEvent(datetime(2019, 1, 1, 15, 59, 30)),
                MarketEvent(datetime(2019, 1, 1, 15, 59, 31)),
                MarketEvent(datetime(2019, 1, 1, 15, 59, 59)),
                MarketEvent(datetime(2019, 1, 1, 16, 0))
            ],
            datetime(2019, 1, 2, 15, 59): [
                MarketEvent(datetime(2019, 1, 1, 16, 0, 1)),
                MarketEvent(datetime(2019, 1, 1, 16, 0, 29)),
                MarketEvent(datetime(2019, 1, 1, 16, 0, 30)),
                MarketEvent(datetime(2019, 1, 1, 16, 0, 31)),
                MarketEvent(datetime(2019, 1, 2, 15, 58, 59)),
                MarketEvent(datetime(2019, 1, 2, 15, 59)),
            ],
            datetime(2019, 1, 2, 16, 0): [
                MarketEvent(datetime(2019, 1, 2, 15, 59, 1)),
                MarketEvent(datetime(2019, 1, 2, 15, 59, 29)),
                MarketEvent(datetime(2019, 1, 2, 15, 59, 30)),
                MarketEvent(datetime(2019, 1, 2, 15, 59, 31)),
                MarketEvent(datetime(2019, 1, 2, 15, 59, 59)),
                MarketEvent(datetime(2019, 1, 2, 16, 0)),
            ]}

    def test_create_partitions_with_latency(self):
        transmitter = Transmitter(
            timesteps=[
                # Day 1.
                datetime(2019, 1, 1, 15, 59),
                datetime(2019, 1, 1, 16),
                # Day 2.
                datetime(2019, 1, 2, 15, 59),
                datetime(2019, 1, 2, 16),
            ])
        events = list()
        for t in transmitter.timesteps:
            for delay in [-1, 0, 1, 29, 30, 31]:
                events.append(MarketEvent(t+timedelta(seconds=delay)))
        transmitter.add_events(events)
        transmitter._create_partitions(latency=30)
        assert transmitter._partition_nonlatent == {
            datetime(2019, 1, 1, 15, 59): [
                MarketEvent(datetime(2019, 1, 1, 15, 58, 59)),
                MarketEvent(datetime(2019, 1, 1, 15, 59))],
            datetime(2019, 1, 1, 16, 0): [
                #MarketEvent(datetime(2019, 1, 1, 15, 59, 1)),
                #MarketEvent(datetime(2019, 1, 1, 15, 59, 29)),
                #MarketEvent(datetime(2019, 1, 1, 15, 59, 30)),
                MarketEvent(datetime(2019, 1, 1, 15, 59, 31)),
                MarketEvent(datetime(2019, 1, 1, 15, 59, 59)),
                MarketEvent(datetime(2019, 1, 1, 16, 0))
            ],
            datetime(2019, 1, 2, 15, 59): [
                #MarketEvent(datetime(2019, 1, 1, 16, 0, 1)),
                #MarketEvent(datetime(2019, 1, 1, 16, 0, 29)),
                #MarketEvent(datetime(2019, 1, 1, 16, 0, 30)),
                MarketEvent(datetime(2019, 1, 1, 16, 0, 31)),
                MarketEvent(datetime(2019, 1, 2, 15, 58, 59)),
                MarketEvent(datetime(2019, 1, 2, 15, 59)),
            ],
            datetime(2019, 1, 2, 16, 0): [
                #MarketEvent(datetime(2019, 1, 2, 15, 59, 1)),
                #MarketEvent(datetime(2019, 1, 2, 15, 59, 29)),
                #MarketEvent(datetime(2019, 1, 2, 15, 59, 30)),
                MarketEvent(datetime(2019, 1, 2, 15, 59, 31)),
                MarketEvent(datetime(2019, 1, 2, 15, 59, 59)),
                MarketEvent(datetime(2019, 1, 2, 16, 0)),
            ]}
        assert transmitter._partition_latent == {
            #datetime(2019, 1, 1, 15, 59): [
            #    MarketEvent(datetime(2019, 1, 1, 15, 58, 59)),
            #    MarketEvent(datetime(2019, 1, 1, 15, 59))],
            datetime(2019, 1, 1, 16, 0): [
                MarketEvent(datetime(2019, 1, 1, 15, 59, 1)),
                MarketEvent(datetime(2019, 1, 1, 15, 59, 29)),
                MarketEvent(datetime(2019, 1, 1, 15, 59, 30)),
                #MarketEvent(datetime(2019, 1, 1, 15, 59, 31)),
                #MarketEvent(datetime(2019, 1, 1, 15, 59, 59)),
                #MarketEvent(datetime(2019, 1, 1, 16, 0))
            ],
            datetime(2019, 1, 2, 15, 59): [
                MarketEvent(datetime(2019, 1, 1, 16, 0, 1)),
                MarketEvent(datetime(2019, 1, 1, 16, 0, 29)),
                MarketEvent(datetime(2019, 1, 1, 16, 0, 30)),
                #MarketEvent(datetime(2019, 1, 1, 16, 0, 31)),
                #MarketEvent(datetime(2019, 1, 2, 15, 58, 59)),
                #MarketEvent(datetime(2019, 1, 2, 15, 59)),
            ],
            datetime(2019, 1, 2, 16, 0): [
                MarketEvent(datetime(2019, 1, 2, 15, 59, 1)),
                MarketEvent(datetime(2019, 1, 2, 15, 59, 29)),
                MarketEvent(datetime(2019, 1, 2, 15, 59, 30)),
                #MarketEvent(datetime(2019, 1, 2, 15, 59, 31)),
                #MarketEvent(datetime(2019, 1, 2, 15, 59, 59)),
                #MarketEvent(datetime(2019, 1, 2, 16, 0)),
            ]}

    @pytest.mark.parametrize('latency, expected_raise', [
        (59, False), (60, True), (61, True),
    ])
    def test_raises_if_latency_greater_than_min_timestep(self, latency, expected_raise):
        transmitter = Transmitter(
            timesteps=[
                # Day 1.
                datetime(2019, 1, 1, 15, 59),
                datetime(2019, 1, 1, 16),
                # Day 2.
                datetime(2019, 1, 2, 15, 59),
                datetime(2019, 1, 2, 16),
            ])
        events = list()
        for t in transmitter.timesteps:
            for delay in [-1, 0, 1, 29, 30, 31]:
                events.append(MarketEvent(t+timedelta(seconds=delay)))
        transmitter.add_events(events)

        if expected_raise:
            with pytest.raises(ValueError):
                transmitter._create_partitions(latency)
        else:
            # Does not raise.
            transmitter._create_partitions(latency)



"""Execution details of a single trade."""
from datetime import datetime
from tradingenv.contracts import AbstractContract, Cash
from tradingenv.broker.fees import IBrokerFees, BrokerFees
import numpy as np


class Trade:
    """Execution details of a single trade. This class is instanced by
    RebalancingResponse.

    Attributes
    ----------
    acq_price : float
        Average acquisition or liquidation price across all lots to execute
        the trade. By default, there is no market impact nor slippage.
        Therefore, avg_price corresponds to the bid price (ask price) in case
        of buy (sell).
    cost_of_cash : float
        Cash to be paid (earned) upfront to buy (sell) the contract.
    cost_of_commissions : float
        Broker commissions paid when transacting.
    """

    __slots__ = (
        "time",
        "contract",
        "quantity",
        "bid_price",
        "ask_price",
        "notional",
        "acq_price",
        "cost_of_cash",
        "cost_of_commissions",
        "cost_of_spread",
    )

    def __init__(
        self,
        time: datetime,
        contract: AbstractContract,
        quantity: float,
        bid_price: float,
        ask_price: float,
        broker_fees: "IBrokerFees" = BrokerFees(),
    ):
        """
        Parameters
        ----------
        time : datetime
            Execution time of the trade.
        contract : AbstractContract
            Traded contract.
        quantity : float
            Traded quantity. A negative number denotes a sell, liquidation or
            short-selling.
        bid_price : float
            Bid price in the limit order book of the contract. Bid price must
            be non-greater then the ask price.
        ask_price : float
            Ask price in the limit order book of the contract. Ask price must
            be non.smaller than the bid price.
        broker_fees : IBrokerFees
            An concrete implementation of AbstractBrokerFees,
            responsible to calculate the total broker fees to be paid
            for the trade.
        """
        if np.isnan(bid_price):
            raise ValueError("Missing bid price for contract {}.".format(contract))
        if np.isnan(ask_price):
            raise ValueError("Missing ask price for contract {}.".format(contract))
        if np.isnan(quantity):
            raise ValueError("Missing quantity for contract {}.".format(contract))
        if quantity == 0:
            raise ValueError("Quantity for contract {} is zero.".format(contract))
        if isinstance(contract, Cash):
            raise ValueError("Contract Cash cannot be traded.")
        self.time = time
        self.contract = contract
        self.quantity = quantity
        self.bid_price = bid_price
        self.ask_price = ask_price
        self.acq_price = ask_price if quantity > 0 else bid_price
        self.notional = self.acq_price * quantity * contract.multiplier
        self.cost_of_cash = self.notional * contract.cash_requirement
        self.cost_of_commissions = broker_fees.commissions(self)
        self.cost_of_spread = (
            abs(quantity) * contract.multiplier * (ask_price - bid_price)
        )

    def __repr__(self) -> str:
        """e.g. Trade(1991-11-05 00:00:00; Buy 340 ETF(SPY) at 282.16)"""
        signal = "Buy" if self.quantity > 0 else "Sell"
        return "{}({}; {} {} {} at {})" "".format(
            self.__class__.__name__,
            self.time,
            signal,
            abs(self.quantity),
            self.contract,
            self.acq_price,
        )

    def __eq__(self, other: 'Trade'):
        return all([
            self.time == other.time,
            self.contract == other.contract,
            np.isclose(self.quantity, other.quantity),
            self.acq_price == other.acq_price
        ])



from tradingenv.events import Observer, EventNewObservation
from tradingenv.features import Feature
from gym.spaces import Space
from datetime import datetime
from typing import Dict, Any, Sequence, Callable, Union
from collections import deque
import tradingenv
import numpy as np
import functools
import math
import copy
import gym


class IState(Observer):
    """An object storing the state of the environment. If you need
    OpenAI-gym compatibility, you are expected to implement IState.space and
    IState.parse in order to define the appropriate data structure to store
    the state."""
    # TODO: Add support for hierarchical features?
    space: Space = None

    def __new__(cls, *args, **kwargs):
        obj: 'Observer' = super().__new__(cls)
        obj._init_args = args
        obj._init_kwargs = kwargs
        obj._cache = dict()
        obj._nr_callbacks = -1
        obj._cache_enabled = False
        obj.history: Dict[datetime, Any] = None
        obj.exchange: 'tradingenv.exchange.Exchange' = None
        obj.features: Sequence[Feature] = None
        obj.save: bool = True
        obj.current_weights: Callable
        obj.reset()
        return obj

    def __init__(self, features: Sequence[Feature] = None, save: bool = True):
        """
        Parameters
        ----------
        features
            A sequence of classes implementing Feature. For small
            states, you probably don't need this. If your state class is
            becoming too big or if you need features transformation, a common
            requirement in machine learning applications, implementing
            Feature allows logic decoupling as IState will be a
            collection of features. Each feature is an observer of events and
            can define its own logic decoupled from the rest of the state or
            other features.
            Note: if you implement features, you don't have to implement any
            other method of this class. You are expected to implement only
            feature and their abstract methods and attributed. See tests for
            examples.
        """
        # TODO: should 'save' be passed to features to guarantee consistency?
        self.features = features
        self.save = save
        if self.features is not None:
            # TODO: test.
            self._verify_features = True
            self._transform_features = True
            names = [f.name for f in self.features]
            repeated_names = set([x for x in names if names.count(x) > 1])
            if repeated_names:
                raise ValueError(f"Duplicate features: {repeated_names}")
            mapping = {
                feature.name: feature.space for feature in self.features
            }
            if all(space is not None for space in mapping.values()):
                self.space = gym.spaces.Dict(mapping)

    def __call__(self, verify: bool = True):
        """Returns self is self.space is None. Otherwise, an object compatible
        with the space will be returned."""
        try:
            state = self.parse()
        except NotImplementedError:
            state = self
        else:
            if verify and self.space is not None:
                if state not in self.space:
                    raise ValueError(
                        'The following state does not belong to the '
                        'observation space.\n{}'.format(repr(state))
                    )
            if self.save:
                t = self._get_global_last_update()
                if self._cache_enabled:
                    if t not in self._cache:
                        self._cache[t] = state
                    self.history[t] = self._cache[t]
                else:
                    # Copy so that to avoid changing history if parse returns a
                    # mutable object (e.g. dict) that changes with new events.
                    self.history[t] = copy.deepcopy(state)
        return state

    def _get_global_last_update(self):
        """Returns time of the most recent update of the state or its features.
        """
        times = list()
        if self.last_update is not None:
            times.append(self.last_update)
        if self.exchange is not None:
            # TODO: test time is also taken from exchange.
            if self.exchange.last_update is not None:
                times.append(self.exchange.last_update)
        if self.features is not None:
            for feature in self.features:
                if feature.last_update is not None:
                    times.append(feature.last_update)
        # This will raise if mixed time zones.
        # TypeError: can't compare offset-naive and offset-aware datetimes
        now = max(times) if times else None
        return now

    def parse(self):
        """Returns an object compatible with IState.space. If IState.space is
        None, you don't have to worry about implementing this method as
        IState() will simply return self. If features are provided, no need
        to implement IState.parse."""
        if self.features is not None:
            # TODO: test feature() and not feature.parse() so that saved in history.
            return {
                feature.name: feature(
                    verify=self._verify_features,
                    transform=self._transform_features
                )
                for feature in self.features
            }
        raise NotImplementedError(
            "If State.space is an instance of gym.spaces.Space (i.e. not "
            "None), then you are supposed to implement IState.parse, which is "
            "expected to return the current state in a format such that "
            "'object in State.space' is True. Note that: specifying "
            "State.space is:\n"
            "- Optional by default.\n"
            "- Required if you want to validate if the state belongs to the "
            "observation space.\n"
            "- Required if you want compatibility with gym spaces "
            "(i.e. IState.space is not None).\n"
            "- Required if you want to use @cache.\n"
            "- Unnecessary if you have provided a list of features during init."
        )

    def reset(
            self,
            exchange: 'tradingenv.exchange.Exchange' = None,
            action_space: 'tradingenv.spaces.PortfolioSpace' = None,
            broker: 'tradingenv.broker.Broker' = None,
    ):
        """Reset state and features for next episode."""
        super().reset()
        self.history = dict()
        self.exchange = exchange
        self.action_space = action_space
        self.broker = broker
        if self.features is not None:
            for feature in self.features:
                feature.reset(exchange, action_space, broker)


class State(IState):
    """This State implementation allows to specify a window of past data. Useful
    when the observation is not Markovian."""

    def __init__(
            self,
            features: Union[int, Sequence[str]],
            window: int = 1,
            stride: int = None,
            max_: float = 5.,
    ):
        """
        Parameters
        ----------
        features
            Number of features or list of feature names. If int, feature names
            will be 0, 1, ..., n-1.
        window
            Window size of the observations. In other words, the observation
            returned by the environment in a given timestep is given by the
            concatenation of as many past observations. Using a window larger
            than 1 can be useful when past observation may have informational
            value that is observations are not Markovian.
        max_
            Maximum value of each feature. This is used to scale the state.
            The state is scaled to [-max_, +max_]. The default value is 100,
            arguably a large value. This is because we defer sanity checks to
            input features to data preprocessing.
        """
        try:
            list(features)
        except TypeError:
            n = features
            self.names = list(range(n))
        else:
            n = len(features)
            self.names = list(features)
        m = window if stride is None else math.ceil(window / stride)
        self.space = gym.spaces.Box(-max_, max_, (m, n), float)
        self.queue = deque(maxlen=window)
        self.last_event = None
        self.stride = stride
        super().__init__()

    def process_EventNewObservation(self, event: EventNewObservation):
        if self.last_event is None:
            for _ in range(self.queue.maxlen):
                self.queue.append([event.to_list()])
        self.queue.append([event.to_list()])
        self.last_event = event

    def parse(self) -> np.array:
        """Returns a numpy array of shape (window, n_features)."""
        x = np.concatenate(self.queue)
        if self.stride:
            # Select a row each self.skip_frames starting from the end.
            # [::-1] flips back down sampled array.
            x = x[::-self.stride][::-1]
        return x

    def flatten(self) -> dict:
        """Flatten the history of past states. Can be useful to parse complex
        states in tabular form for further analysis or visualisations."""
        s_ = dict()
        for k, v in self.history.items():
            try:
                if v.ndim == 1:
                    # To make this work with 1d array
                    s_[k] = v.item()
                elif v.ndim == 2:
                    # To make this work with 2d array
                    s_[k] = v.squeeze()
            except ValueError:
                # Was not a np.array, either 1d or 2d.
                # ValueError: can only convert an array of size 1 to a Python scalar
                for i, j in enumerate(v[0]):
                    s_[f'{k}{i}'] = j
                pass
        return s_


def cache(callback=None, check_every: int = np.inf, check_on_start: bool = True):
    """Decorator which allows to use cached results from previous
    episodes/backtests of Feature.process_<Event> methods to save computations.

    Parameters
    ----------
    callback
        A convenience arg all allows cache being either callable or not.
    check_every : bool
        If True, every 'check_every' the cached result will not be used and
        compared with the cached result. If there is basis_naive_tm1 difference, an error
        will be raised, meaning that your features should probably not used
        cached results. See Notes.
    check_on_start : bool
        If True, cached and non-cached results will be compared when the
        feature will receive its first observed event. If they are different,
        an error will be raised, meaning that your features should probably
        not used cached results. See Notes.

    Notes
    -----
    By cached result we mean the state of the feature calculated at the same
    timestep from basis_naive_tm1 previous episode. By the nature of this caching mechanism,
    there is no way to use cached results when your feature is not deterministic
    and in such case an error will be raised presumably at the first check,
    assuming that some form of checking is performed as indicated by the
    arguments of this function.
    """
    # Implementation inspired by 2018 update of jsbueno:
    #   https://stackoverflow.com/questions/10294014
    if callback is None:
        # This trick allows to use either calling this decorator or not.
        # If you are here, your decorator has been called with kwargs.
        return functools.partial(
            cache,
            check_every=check_every,
            check_on_start=check_on_start,
        )

    def are_equal(a, b):
        if isinstance(a, dict):
            return all(are_equal(a[k], b[k]) for k in a)
        return np.allclose(a, b, equal_nan=True)

    @functools.wraps(callback)
    def wrapper(self: IState, event: 'tradingenv.IEvent'):
        self._cache_enabled = True
        verify = any([
            self._nr_callbacks > 0 and self._nr_callbacks % check_every == 0,
            check_on_start and self._nr_callbacks == 0,
        ])
        if event.time in self._cache:
            actual = self._cache[event.time]
            if verify:
                callback(self, event)
                expected = self.parse()
                if not are_equal(expected, actual):
                    raise ValueError(
                        'Chached data is different from what the state should '
                        'be. You should not use @cache if the state is '
                        'path-dependant:\n'
                        'Actual:\n'
                        '{actual}\n'
                        'Expected:\n'
                        '{expected}'
                        ''.format(actual=actual, expected=expected)
                    )
        else:
            callback(self, event)
            actual = self.parse()
        return actual
    return wrapper


import setuptools
import os
here = os.path.abspath(os.path.dirname(__file__))

# Parse __version__ and __package__. Reference for why it's done that way:
#   https://stackoverflow.com/questions/2058802
with open(os.path.join(here, "tradingenv", "__init__.py"), encoding="utf-8") as f:
    init_file_lines = f.readlines()
for line in init_file_lines:
    if line.startswith(("__package__", "__version__")):
        exec(line.rstrip())

# long_description
with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# package_data.
package_data_paths = list()
for root, dirs, files in os.walk(os.path.join(here, "tradingenv")):
    for file in files:
        if file.endswith(".csv") or file.endswith(".txt"):
            path = os.path.join(root, file)
            package_data_paths.append(path)

setuptools.setup(
    name=__package__,
    version=__version__,
    description="Backtest trading strategies or train reinforcement learning "
                "agents with and event-driven market simulator.",
    long_description=long_description,
    url="https://github.com/xaiassetmanagement/tradingenv",
    author="Federico Fontana",
    author_email="federico.fontana@xai-am.com",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Office/Business :: Financial :: Investment",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
    ],
    keywords=["trading", "investment", "finance", "backtest", "reinforcement-learning", "gym"],
    packages=setuptools.find_packages(exclude=["docs", "notebooks", "tests"]),
    install_requires=[
        "exchange_calendars",
        "gym==0.21",
        "numpy",
        "pandas",
        "pandas_market_calendars",
        "plotly",
        "scikit-learn",
        "statsmodels",
        "tqdm",
    ],
    extras_require={
        "extra": [
            "yfinance",
            "pytest",
            "pytest-mock",
            "pytest-cov",
            "openpyxl",
            "jupyterlab",
            "nbconvert",  # https://stackoverflow.com/questions/65376052
            "Sphinx",
        ],
    },
    package_data={"tradingenv": package_data_paths},
)



"""TradingEnv is an even-driven market simulator. All custom events must
inherit the class IEvent."""
import tradingenv
from tradingenv.broker import broker
from typing import Sequence
import numpy as np
from datetime import datetime
from abc import ABC
from typing import Dict, Mapping
import inspect


class Observer(ABC):
    """
    Attributes
    ----------
    _observed_events : Sequence[Type['tradingenv.events.IEvent']]
        A sequence of class Events.  The state of a Feature is updated
        automatically by TradingEnv whenever an observed event occurs
        (i.e. it is passed to TradingEnv.add_event). By default, no event
        are observed.

    Notes
    -----
    Generally in the observer patterns we store the functions to be run
    whenever an observed event is processed. Here we register the method id
    instead, which is slightly slower and certainly less elegant. The reason
    is that deepcopies of the env will introduce bucks because features will
    be copied but not their callback methods from the observed_events, which
    might introduce nasty bugs (e.g. when using ray).

    Ideas for alternative implementations of the observer pattern:
        https://stackoverflow.com/questions/1092531/event-system-in-python
    """

    def __new__(cls, *args, **kwargs) -> 'Observer':
        observer = super().__new__(cls)
        observer.last_update: datetime = None
        observer._nr_callbacks = 0
        observer.name = cls.__name__
        observer._observed_events = observer._get_observed_events()
        observer._init_args = args
        observer._init_kwargs = kwargs
        return observer

    def __call__(self, *args, **kwargs):
        """Called as soon as the Observer has finished processing a whatever
        event. Implement if you need to add extra logic (e.g. checkpoint in
        history current state of the observer/feature/state."""

    def reset(self):
        """Re-initialise from scratch, discard attributes that don't come
        with the initialisation."""
        # TODO: test
        self.__init__(*self._init_args, **self._init_kwargs)
        self.last_update = None
        self._nr_callbacks = 0

    def _get_observed_events(self) -> Dict[str, str]:
        """Infer observed events from class methods starting with 'process_'."""
        observed_events = dict()
        for attr_name in dir(self.__class__):
            if attr_name.startswith("process_"):
                event_name = attr_name.replace("process_", "")
                self._verify_callback_signature(event_name)
                observed_events[event_name] = attr_name
        return observed_events

    def _verify_is_observing(self):
        if len(self._observed_events) == 0:
            raise AttributeError(
                "You must implement at least one {}.process_<EventName>"
                "".format(self.name)
            )

    def _verify_callback_signature(self, event_name: str):
        """Verify that callback methods accept argument 'event' only."""
        callback_name = "process_{}".format(event_name)
        callback = getattr(self, callback_name)
        if list(inspect.signature(callback).parameters) != ["event"]:
            raise TypeError(
                "{}.{} must accept the single argument 'event'."
                "".format(self.__class__.__name__, callback_name)
            )


class IEvent:
    """TradingEnv is an even-driven simulator where events are passed using
    TradingEnv.add_event. All events must inherit this class.

    Attributes
    ----------
    time : datetime
        Timestamp associated with the event, i.e. when the event occurred.
    """
    time: datetime = None  # class attribute set during instance initialization

    def __repr__(self) -> str:
        return "{}({})".format(self.__class__.__name__, self.time)

    def __le__(self, other: "IEvent") -> bool:
        """Events are sorted by time."""
        return self.time <= other.time

    def __lt__(self, other: "IEvent") -> bool:
        """Events are sorted by time."""
        return self.time < other.time

    def __eq__(self, other: "IEvent") -> bool:
        """Two events are equal if and only if all attributes are equal."""
        return self._public_attr(self) == self._public_attr(other)

    def _public_attr(self, obj: object) -> dict:
        """Returns vars(obj) dropping attributes starting with '_' this
        representing private attributes."""
        return {k: v for k, v in vars(obj).items() if not k.startswith("_")}

    def notify(self, observers: Sequence['tradingenv.events.Observer']):
        for observer in observers:
            if type(self).__name__ in observer._observed_events:
                callback_name = observer._observed_events[type(self).__name__]
                callback = getattr(observer, callback_name)
                callback(self)
                observer.last_update = self.time
                observer._nr_callbacks += 1
                observer()


class EventNBBO(IEvent):
    """(Synchronous) National Best Bid and Offer."""

    __slots__ = ("contract", "bid_price", "ask_price", "bid_size", "ask_size", "time")

    def __init__(
        self,
        time: datetime,
        contract: "tradingenv.contracts.AbstractContract",
        bid_price: float,
        ask_price: float,
        bid_size: float = np.inf,
        ask_size: float = np.inf,
    ):
        """
        Parameters
        ----------
        contract : 'tradingenv.contracts.AbstractContract'
            Contract ID. Generally a string (e.g. 'S&P 500') but could be a
            whatever hashable object.
        bid_price : float
            Bid transaction_price now.
        ask_price : float
            Ask transaction_price now.
        bid_size : float
            Bid size now (top of the book).
        ask_size : float
            Ask size now (top of the book).
        time : datetime
            Timestamp associated with the NBBO. If not provided, the time will
            be assumed to be the current UTC time.
        """
        self.time = time
        self.contract = contract
        self.bid_price = bid_price
        self.ask_price = ask_price
        self.mid_price = (bid_price + ask_price) / 2
        self.bid_size = bid_size
        self.ask_size = ask_size
        contract.verify(self.mid_price)

    def __repr__(self) -> str:
        return repr({attr: getattr(self, attr) for attr in self.__slots__})

    def __eq__(self, other: "EventNBBO") -> bool:
        return all(
            getattr(self, attr) == getattr(other, attr)
            for attr in self.__slots__
        )


class EventContractDiscontinued(IEvent):
    def __init__(self, time: datetime, contract: "tradingenv.contracts.AbstractContract"):
        self.time = time
        self.contract = contract


class EventReset(IEvent):
    """IEvent signaling that TradingEnv.reset has finished to run."""

    def __init__(self, time: datetime, track_record: "broker.TrackRecord"):
        self.time = time
        self.track_record = track_record


class EventStep(IEvent):
    """IEvent signaling that TradingEnv.step has finished to run."""

    def __init__(self, time: datetime, track_record: "broker.TrackRecord", action):
        self.time = time
        self.track_record = track_record
        self.action = action


class EventDone(IEvent):
    """IEvent signaling that TradingEnv.step has finished to run."""
    def __init__(self, time: datetime, broker: "broker.Broker"):
        self.time = time
        self.broker = broker


class EventNewDate(IEvent):
    """Triggered just before the first event of the date is processed."""
    def __init__(self, time: datetime, broker: "broker.Broker"):
        self.time = time
        self.broker = broker


class EventNewObservation(IEvent):
    """Stores a snapshot of exogenous variables at a given time."""

    def __init__(self, time: datetime, data: Mapping):
        self.time = time
        self.data = dict(data)

    def __getitem__(self, item):
        return self.data[item]

    def __len__(self):
        return len(self.data)

    def to_list(self) -> np.array:
        x = list(self.data.values())
        return x



import tradingenv
from tradingenv.events import Observer
from sklearn.exceptions import NotFittedError
from datetime import datetime
from typing import Dict, Any
import numpy as np
import gym.spaces
import copy


class Feature(Observer):
    """This class gives you the option to instance State as a collection of
    features. See docstring of State for further details.

    Attributes
    ----------
    exchange
        Instance of tradingenv.exchange.Exchange, storing current and
        historical asset prices.
    action_space
        Action space passed when instancing the environemtn.
    broker
        Instance of tradingenv.broker.Broker storing current holdings,
        net liquidation value, pnl, past track record, past rebalancing
        requests, past trades, past commissions and more.
    """
    # Type hinting here, set in Feature.reset
    exchange: 'tradingenv.exchange.Exchange' = None
    action_space: 'tradingenv.spaces.PortfolioSpace' = None
    broker: 'tradingenv.broker.Broker' = None

    def __init__(
            self,
            space: gym.spaces.Space = None,
            name: str = None,
            save: bool = True,
            transformer = None,
    ):
        """
        Parameters
        ----------
        space
            An optional Space needed for OpenAI-gym compatibility. Note:
            only the feature before transformations (if any) will be validated
            against the space.
        name
            An optional name of the feature, class name by default. It is
            useful to provide a custom name to differentiate the same feature
            in the state if provided more than once e.g. with different
            parameters.
        save
            If True (default), the output of Feature.parse() will be
            automatically saved to feature.history whenever (1) an observed
            event is processed by the feature or (2) feature() is called.
        transformer
            A sklearn preprocessing transformer, defaults to None. If provided,
            by default the features returned by parse will be transformed.
            The easiest way to fitting transformers is to pass
            fit_transformer=True when instancing TradingEnv. You can set
            custom routines either by passing extending transformers or by
            implementing Feature._manual_fit_transformer.
        """
        # Note: optionally provide support for transformer.fit_partial as a
        # 'streaming' fit. Beware of double counting depending on the approach
        # used to call .fit_partial.
        if space is None and transformer is not None:
            raise ValueError(
                "You must provide a 'space' in order to use 'transformer'."
            )
        self.space = space
        self.name = name or self.name
        self.transformer = transformer
        self.save = save
        self.history: Dict[datetime, Any] = dict()

    def __call__(self, verify: bool = False, transform: bool = True):
        """
        Parameters
        ----------
        verify
            If False (default), the (un)transformed feature if validated
            against the space. The transformed feature is never validated
            against the transformed space as it's responsibility of the
            transformer to run further checks if any. In other words, it's
            possible for a transformed feature to fall outside the transformed
            space without raising errors.
            Here obs does not belong to space but calling feature() does not
            raise by default. This is by design for two reasons:
            (1) big features might not belong to the space until they warm up
            (2) state is verify=True by default, so avoid double verification.
        transform
            If True (default) and if a transformer is provided during
            initialisation of this class, then the feature will be transformed.

        Returns
        -------
        Parsed feature compatible with Feature.space, either transformed or not.
        If Feature.parse has not been implemented, returns self.
        """
        # TODO: you are not using 'transform' input.
        # TODO: when warming up batch transformer, transform must be = False
        # TODO: if transformation occurs, what do we save to history? Test.
        try:
            feature = self.parse()
        except NotImplementedError:
            feature = self
        else:
            verify = verify and self.space is not None
            if verify and (feature not in self.space):
                raise ValueError(
                    'The following state does not belong to the observation '
                    'space.\n{}'.format(repr(feature))
                )
            try:
                # 1.79 µs ± 10.5 ns per loop
                feature = self.transformer.transform(feature)
            except (NotFittedError, AttributeError):
                # AttributeError: 'NoneType' object has no attribute 'transform'
                # Copy so that to avoid changing history if parse returns a
                # mutable object (e.g. dict) that changes with new events.
                feature = copy.deepcopy(feature)
            finally:
                self._save_observation(feature)
        return feature

    def _now(self):
        """If the feature is not observing events but parsing some data from
        broker or exchange, feature.last_update will be none will be none
        while time is progressing. That is the reason why we seek if time
        has progressed in other parts of the environment. This is also a strong
        red flag that we would need a centralised Clock in the simulator."""
        # TODO: test.
        now = self.last_update
        if self.exchange is not None:
            if self.exchange.last_update is not None:
                now = max(now or datetime.min, self.exchange.last_update)
        return now

    def reset(
            self,
            exchange: 'tradingenv.exchange.Exchange' = None,
            action_space: 'tradingenv.spaces.PortfolioSpace' = None,
            broker: 'tradingenv.broker.Broker' = None,
    ):
        """Reset the feature to its original state. We pass exchange,
        action_space and broker as they could be needed to compute some features
        such as current positions, portfolio weights, market prices etc."""
        # TODO: Test.
        _transfomer = self.transformer
        super().reset()
        # Make the transformer persistent across resets.
        self.transformer = _transfomer
        self.exchange = exchange
        self.action_space = action_space
        self.broker = broker

    def _save_observation(self, obs):
        # TODO: history could be extended to class History(dict) implementing
        # extra methods to parse and visualise.
        if self.save:
            self.history[self._now()] = obs

    def _parse_history(self, make_2d: bool = False):
        """Parse history of the feature in a format that can be used to
        fit the transformer."""
        if issubclass(type(self.space), gym.spaces.Box):
            # If space shape is (x, y, z, ...), shape of data will
            # be (n, x, y, z, ...).
            if len(self.history) == 0:
                raise ValueError(
                    f"You are trying to fit the transformer of feature "
                    f"{self.name} on a empty history. The easiest way to "
                    f"fix this is to pass fit_transformers=True when "
                    f"instancing TradingEnv."
                )
            history = list(self.history.values())
            data = np.concatenate([history])
            if make_2d:
                # Transformers expect 2D data. Meaning that (m, n) shaped
                # observations should be flattened
                # ValueError: Expected 2D array, got XD array instead
                # This operation transforms features from (nr_obs, m, n) to
                # (nr_obs, m * n).
                n, rows, cols = data.shape
                # https://stackoverflow.com/questions/53870113
                # ValueError: Found array with dim 3. StandardScaler expected <= 2.
                # So we squeeze
                size_flattened_feature = rows * cols
                data = data.reshape(n, size_flattened_feature)
        elif issubclass(type(self.space), gym.spaces.MultiBinary):
            # No need to parse history as there is no need to fit a transformer
            # on dummy variables.
            data = None
        else:
            raise NotImplementedError(
                "You are attempting to fit a transformer in an unsupported "
                "space: {self.space}. A space must be supported to "
                "retrieve batch observations to fit. Use a support space "
                "of implement Feature._manual_fit_transformer."
                "Supported spaces are: {gym.spaces.Box}."
            )
        return data

    def fit_transformer(self):
        """Fit transformed using all historical observations (batch) by default.
        The user can optionally implement _manual_fit_transformer to
        override the procedure."""
        try:
            self._manual_fit_transformer()
        except NotImplementedError:
            data = self._parse_history(make_2d=True)
            self.transformer.fit(data)

    def parse(self):
        """Returns any data structure representing the current value assumed
        by the feature. If a 'space' is provided when instancing this Feature,
        the returned value will be validated against the space if verify=True.
        This method is also required if you desire to store historical values
        of the feature in Feature.history whenever an observed event is
        received."""
        raise NotImplementedError()

    def _manual_fit_transformer(self):
        """You can implement here custom procedures to fit the transformer
        when Feature.fit_transformer is called."""
        raise NotImplementedError()
